#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import torch
import numpy as np
from pathlib import Path
from matplotlib import pyplot as plt
import torch.utils.data as Data
from typing import Optional
from vame.schemas.states import EvaluateModelFunctionSchema, save_state
from vame.util.auxiliary import read_config
from vame.model.rnn_vae import RNN_VAE
from vame.model.dataloader import SEQUENCE_DATASET
from vame.logging.logger import VameLogger


logger_config = VameLogger(__name__)
logger = logger_config.logger

use_gpu = torch.cuda.is_available()
if use_gpu:
    pass
else:
    torch.device("cpu")

def plot_reconstruction(
    filepath: str,
    test_loader: Data.DataLoader,
    seq_len_half: int,
    model: RNN_VAE,
    model_name: str,
    FUTURE_DECODER: bool,
    FUTURE_STEPS: int,
    suffix: Optional[str] = None
) -> None:
    """
    Plot the reconstruction and future prediction of the input sequence.

    Args:
        filepath (str): Path to save the plot.
        test_loader (Data.DataLoader): DataLoader for the test dataset.
        seq_len_half (int): Half of the temporal window size.
        model (RNN_VAE): Trained VAE model.
        model_name (str): Name of the model.
        FUTURE_DECODER (bool): Flag indicating whether the model has a future prediction decoder.
        FUTURE_STEPS (int): Number of future steps to predict.
        suffix (Optional[str], optional): Suffix for the saved plot filename. Defaults to None.
    """
    #x = test_loader.__iter__().next()
    dataiter = iter(test_loader)
    x = next(dataiter)
    x = x.permute(0,2,1)
    if use_gpu:
        data = x[:,:seq_len_half,:].type('torch.FloatTensor').cuda()
        data_fut = x[:,seq_len_half:seq_len_half+FUTURE_STEPS,:].type('torch.FloatTensor').cuda()
    else:
        data = x[:,:seq_len_half,:].type('torch.FloatTensor').to()
        data_fut = x[:,seq_len_half:seq_len_half+FUTURE_STEPS,:].type('torch.FloatTensor').to()
    if FUTURE_DECODER:
        x_tilde, future, latent, mu, logvar = model(data)

        fut_orig = data_fut.cpu()
        fut_orig = fut_orig.data.numpy()
        fut = future.cpu()
        fut = fut.detach().numpy()

    else:
        x_tilde, latent, mu, logvar = model(data)

    data_orig = data.cpu()
    data_orig = data_orig.data.numpy()
    data_tilde = x_tilde.cpu()
    data_tilde = data_tilde.detach().numpy()

    if FUTURE_DECODER:
        fig, axs = plt.subplots(2, 5)
        fig.suptitle('Reconstruction [top] and future prediction [bottom] of input sequence')
        for i in range(5):
            axs[0,i].plot(data_orig[i,...], color='k', label='Sequence Data')
            axs[0,i].plot(data_tilde[i,...], color='r', linestyle='dashed', label='Sequence Reconstruction')

            axs[1,i].plot(fut_orig[i,...], color='k')
            axs[1,i].plot(fut[i,...], color='r', linestyle='dashed')
        axs[0,0].set(xlabel='time steps', ylabel='reconstruction')
        axs[1,0].set(xlabel='time steps', ylabel='predction')
        fig.savefig(os.path.join(filepath,"evaluate",'Future_Reconstruction.png'))

    else:
        fig, ax1 = plt.subplots(1, 5)
        for i in range(5):
            fig.suptitle('Reconstruction of input sequence')
            ax1[i].plot(data_orig[i,...], color='k', label='Sequence Data')
            ax1[i].plot(data_tilde[i,...], color='r', linestyle='dashed', label='Sequence Reconstruction')
        fig.set_tight_layout(True)
        if not suffix:
            fig.savefig(os.path.join(filepath,'evaluate','Reconstruction_'+model_name+'.png'), bbox_inches='tight')
        elif suffix:
            fig.savefig(os.path.join(filepath,'evaluate','Reconstruction_'+model_name+'_'+suffix+'.png'), bbox_inches='tight')


def plot_loss(cfg: dict, filepath: str, model_name: str) -> None:
    """
    Plot the losses of the trained model.

    Args:
        cfg (dict): Configuration dictionary.
        filepath (str): Path to save the plot.
        model_name (str): Name of the model.
    """
    basepath = os.path.join(cfg['project_path'],"model","model_losses")
    train_loss = np.load(os.path.join(basepath,'train_losses_'+model_name+'.npy'))
    test_loss = np.load(os.path.join(basepath,'test_losses_'+model_name+'.npy'))
    mse_loss_train = np.load(os.path.join(basepath,'mse_train_losses_'+model_name+'.npy'))
    mse_loss_test = np.load(os.path.join(basepath,'mse_test_losses_'+model_name+'.npy'))
    km_losses = np.load(os.path.join(basepath,'kmeans_losses_'+model_name+'.npy'))
    kl_loss = np.load(os.path.join(basepath,'kl_losses_'+model_name+'.npy'))
    fut_loss = np.load(os.path.join(basepath,'fut_losses_'+model_name+'.npy'))

    fig, (ax1) = plt.subplots(1, 1)
    fig.suptitle('Losses of our Model')
    ax1.set(xlabel='Epochs', ylabel='loss [log-scale]')
    ax1.set_yscale("log")
    ax1.plot(train_loss, label='Train-Loss')
    ax1.plot(test_loss, label='Test-Loss')
    ax1.plot(mse_loss_train, label='MSE-Train-Loss')
    ax1.plot(mse_loss_test, label='MSE-Test-Loss')
    ax1.plot(km_losses, label='KMeans-Loss')
    ax1.plot(kl_loss, label='KL-Loss')
    ax1.plot(fut_loss, label='Prediction-Loss')
    ax1.legend()
    fig.savefig(os.path.join(filepath,"evaluate",'MSE-and-KL-Loss'+model_name+'.png'))


def eval_temporal(
    cfg: dict,
    use_gpu: bool,
    model_name: str,
    fixed: bool,
    snapshot: Optional[str] = None,
    suffix: Optional[str] = None
) -> None:
    """
    Evaluate the temporal aspects of the trained model.

    Args:
        cfg (dict): Configuration dictionary.
        use_gpu (bool): Flag indicating whether to use GPU for evaluation.
        model_name (str): Name of the model.
        fixed (bool): Flag indicating whether the data is fixed or not.
        snapshot (Optional[str], optional): Path to the model snapshot. Defaults to None.
        suffix (Optional[str], optional): Suffix for the saved plot filename. Defaults to None.
    """
    SEED = 19
    ZDIMS = cfg['zdims']
    FUTURE_DECODER = cfg['prediction_decoder']
    TEMPORAL_WINDOW = cfg['time_window']*2
    FUTURE_STEPS = cfg['prediction_steps']
    NUM_FEATURES = cfg['num_features']
    if not fixed:
        NUM_FEATURES = NUM_FEATURES - 2
    TEST_BATCH_SIZE = 64
    hidden_size_layer_1 = cfg['hidden_size_layer_1']
    hidden_size_layer_2 = cfg['hidden_size_layer_2']
    hidden_size_rec = cfg['hidden_size_rec']
    hidden_size_pred = cfg['hidden_size_pred']
    dropout_encoder = cfg['dropout_encoder']
    dropout_rec = cfg['dropout_rec']
    dropout_pred = cfg['dropout_pred']
    softplus = cfg['softplus']

    filepath = os.path.join(cfg['project_path'],"model")


    seq_len_half = int(TEMPORAL_WINDOW/2)
    if use_gpu:
        torch.cuda.manual_seed(SEED)
        model = RNN_VAE(TEMPORAL_WINDOW,ZDIMS,NUM_FEATURES,FUTURE_DECODER,FUTURE_STEPS, hidden_size_layer_1,
                        hidden_size_layer_2, hidden_size_rec, hidden_size_pred, dropout_encoder,
                        dropout_rec, dropout_pred, softplus).cuda()
        model.load_state_dict(torch.load(os.path.join(cfg['project_path'],"model","best_model",model_name+'_'+cfg['Project']+'.pkl')))
    else:
        model = RNN_VAE(TEMPORAL_WINDOW,ZDIMS,NUM_FEATURES,FUTURE_DECODER,FUTURE_STEPS, hidden_size_layer_1,
                        hidden_size_layer_2, hidden_size_rec, hidden_size_pred, dropout_encoder,
                        dropout_rec, dropout_pred, softplus).to()
        if not snapshot:
            model.load_state_dict(torch.load(os.path.join(cfg['project_path'],"model","best_model",model_name+'_'+cfg['Project']+'.pkl'), map_location=torch.device('cpu')))
        elif snapshot:
            model.load_state_dict(torch.load(snapshot), map_location=torch.device('cpu'))
    model.eval() #toggle evaluation mode

    testset = SEQUENCE_DATASET(
        os.path.join(cfg["project_path"], "data", "train", ""),
        data="test_seq.npy",
        train=False,
        temporal_window=TEMPORAL_WINDOW,
        logger_config=logger_config,
    )
    test_loader = Data.DataLoader(testset, batch_size=TEST_BATCH_SIZE, shuffle=True, drop_last=True)

    if not snapshot:
        plot_reconstruction(filepath, test_loader, seq_len_half, model, model_name, FUTURE_DECODER, FUTURE_STEPS)#, suffix=suffix
    elif snapshot:
        plot_reconstruction(filepath, test_loader, seq_len_half, model, model_name, FUTURE_DECODER, FUTURE_STEPS, suffix=suffix)#,
    if use_gpu:
        plot_loss(cfg, filepath, model_name)
    else:
        plot_loss(cfg, filepath, model_name)
        # pass #note, loading of losses needs to be adapted for CPU use #TODO

@save_state(model=EvaluateModelFunctionSchema)
def evaluate_model(config: str, use_snapshots: bool = False, save_logs: bool = False) -> None:
    """Evaluate the trained model.

    Args:
        config (str): Path to config file.
        use_snapshots (bool, optional): Whether to plot for all snapshots or only the best model. Defaults to False.
    """
    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'evaluate_model.log'
            logger_config.add_file_handler(log_path)

        model_name = cfg['model_name']
        fixed = cfg['egocentric_data']

        if not os.path.exists(os.path.join(cfg['project_path'],"model","evaluate")):
            os.mkdir(os.path.join(cfg['project_path'],"model","evaluate"))

        use_gpu = torch.cuda.is_available()
        if use_gpu:
            logger.info("Using CUDA")
            logger.info('GPU active: {}'.format(torch.cuda.is_available()))
            logger.info('GPU used: {}'.format(torch.cuda.get_device_name(0)))
        else:
            torch.device("cpu")
            logger.info("CUDA is not working, or a GPU is not found; using CPU!")

        logger.info("Evaluation of %s model. " %model_name)
        if not use_snapshots:
            eval_temporal(cfg, use_gpu, model_name, fixed)#suffix=suffix
        elif use_snapshots:
            snapshots=os.listdir(os.path.join(cfg['project_path'],'model','best_model','snapshots'))
            for snap in snapshots:
                fullpath = os.path.join(cfg['project_path'],"model","best_model","snapshots",snap)
                epoch=snap.split('_')[-1]
                eval_temporal(cfg, use_gpu, model_name, fixed, snapshot=fullpath, suffix='snapshot'+str(epoch))

        logger.info("You can find the results of the evaluation in '/Your-VAME-Project-Apr30-2020/model/evaluate/' \n"
            "OPTIONS:\n"
            "- vame.pose_segmentation() to identify behavioral motifs.\n"
            "- re-run the model for further fine tuning. Check again with vame.evaluate_model()")
    except Exception as e:
        logger.exception(f"An error occurred during model evaluation: {e}")
    finally:
        logger_config.remove_file_handler()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import numpy as np
from pathlib import Path
import scipy.signal
from scipy.stats import iqr
import matplotlib.pyplot as plt
from typing import List, Optional, Tuple
from vame.logging.logger import VameLogger
from vame.util.auxiliary import read_config
from vame.schemas.states import CreateTrainsetFunctionSchema, save_state
from vame.util.data_manipulation import interpol_all_nans


logger_config = VameLogger(__name__)
logger = logger_config.logger



def plot_check_parameter(
    cfg: dict,
    iqr_val: float,
    num_frames: int,
    X_true: List[np.ndarray],
    X_med: np.ndarray,
) -> None:
    """
    Plot the check parameter - z-scored data and the filtered data.

    Args:
        cfg (dict): Configuration parameters.
        iqr_val (float): IQR value.
        num_frames (int): Number of frames.
        X_true (List[np.ndarray]): List of true data.
        X_med (np.ndarray): Filtered data.
        anchor_1 (int): Index of the first anchor point (deprecated).
        anchor_2 (int): Index of the second anchor point (deprecated).

    Returns:
        None - Plot the z-scored data and the filtered data.
    """
    plot_X_orig = np.concatenate(X_true, axis=0).T
    plot_X_med = X_med.copy()
    iqr_cutoff = cfg['iqr_factor']*iqr_val

    plt.figure()
    plt.plot(plot_X_orig.T)
    plt.axhline(y=iqr_cutoff, color='r', linestyle='--', label="IQR cutoff")
    plt.axhline(y=-iqr_cutoff, color='r', linestyle='--')
    plt.title("Full Signal z-scored")
    plt.legend()

    if num_frames > 1000:
        rnd = np.random.choice(num_frames)

        plt.figure()
        plt.plot(plot_X_med[:,rnd:rnd+1000].T)
        plt.axhline(y=iqr_cutoff, color='r', linestyle='--', label="IQR cutoff")
        plt.axhline(y=-iqr_cutoff, color='r', linestyle='--')
        plt.title("Filtered signal z-scored")
        plt.legend()

        plt.figure()
        plt.plot(plot_X_orig[:,rnd:rnd+1000].T)
        plt.axhline(y=iqr_cutoff, color='r', linestyle='--', label="IQR cutoff")
        plt.axhline(y=-iqr_cutoff, color='r', linestyle='--')
        plt.title("Original signal z-scored")
        plt.legend()

        plt.figure()
        plt.plot(plot_X_orig[:,rnd:rnd+1000].T, 'g', alpha=0.5)
        plt.plot(plot_X_med[:,rnd:rnd+1000].T, '--m', alpha=0.6)
        plt.axhline(y=iqr_cutoff, color='r', linestyle='--', label="IQR cutoff")
        plt.axhline(y=-iqr_cutoff, color='r', linestyle='--')
        plt.title("Overlayed z-scored")
        plt.legend()

        # plot_X_orig = np.delete(plot_X_orig.T, anchor_1, 1)
        # plot_X_orig = np.delete(plot_X_orig, anchor_2, 1)
        # mse = (np.square(plot_X_orig[rnd:rnd+1000, :] - plot_X_med[:,rnd:rnd+1000].T)).mean(axis=0)


    else:
        plt.figure()
        plt.plot(plot_X_med.T)
        plt.axhline(y=iqr_cutoff, color='r', linestyle='--', label="IQR cutoff")
        plt.axhline(y=-iqr_cutoff, color='r', linestyle='--')
        plt.title("Filtered signal z-scored")
        plt.legend()

        plt.figure()
        plt.plot(plot_X_orig.T)
        plt.axhline(y=iqr_cutoff, color='r', linestyle='--', label="IQR cutoff")
        plt.axhline(y=-iqr_cutoff, color='r', linestyle='--')
        plt.title("Original signal z-scored")
        plt.legend()

    logger.info("Please run the function with check_parameter=False if you are happy with the results")

def traindata_aligned(
    cfg: dict,
    files: List[str],
    testfraction: float,
    savgol_filter: bool,
    check_parameter: bool
) -> None:
    """
    Create training dataset for aligned data.

    Args:
        cfg (dict): Configuration parameters.
        files (List[str]): List of files.
        testfraction (float): Fraction of data to use as test data.
        num_features (int): Number of features (deprecated).
        savgol_filter (bool): Flag indicating whether to apply Savitzky-Golay filter.
        check_parameter (bool): If True, the function will plot the z-scored data and the filtered data.

    Returns:
        None - Save numpy arrays with the test/train info to the project folder.
    """

    X_train = []
    pos = []
    pos_temp = 0
    pos.append(0)

    if check_parameter:
        X_true = []
        files = [files[0]]

    for file in files:
        logger.info("z-scoring of file %s" %file)
        path_to_file = os.path.join(cfg['project_path'],"data", file, file+'-PE-seq.npy')
        data = np.load(path_to_file)

        X_mean = np.mean(data,axis=None)
        X_std = np.std(data, axis=None)
        X_z = (data.T - X_mean) / X_std

        # Introducing artificial error spikes
        # rang = [1.5, 2, 2.5, 3, 3.5, 3, 3, 2.5, 2, 1.5]
        # for i in range(num_frames):
        #     if i % 300 == 0:
        #         rnd = np.random.choice(12,2)
        #         for j in range(10):
        #             X_z[i+j, rnd[0]] = X_z[i+j, rnd[0]] * rang[j]
        #             X_z[i+j, rnd[1]] = X_z[i+j, rnd[1]] * rang[j]

        if check_parameter:
            X_z_copy = X_z.copy()
            X_true.append(X_z_copy)

        if cfg['robust']:
            iqr_val = iqr(X_z)
            logger.info("IQR value: %.2f, IQR cutoff: %.2f" %(iqr_val, cfg['iqr_factor']*iqr_val))
            for i in range(X_z.shape[0]):
                for marker in range(X_z.shape[1]):
                    if X_z[i,marker] > cfg['iqr_factor']*iqr_val:
                        X_z[i,marker] = np.nan

                    elif X_z[i,marker] < -cfg['iqr_factor']*iqr_val:
                        X_z[i,marker] = np.nan

            X_z = interpol_all_nans(X_z)

        X_len = len(data.T)
        pos_temp += X_len
        pos.append(pos_temp)
        X_train.append(X_z)

    X = np.concatenate(X_train, axis=0)
    # X_std = np.std(X)

    detect_anchors = np.std(X.T, axis=1)
    sort_anchors = np.sort(detect_anchors)
    if sort_anchors[0] == sort_anchors[1]:
        anchors = np.where(detect_anchors == sort_anchors[0])[0]
        anchor_1_temp = anchors[0]
        anchor_2_temp = anchors[1]

    else:
        anchor_1_temp = int(np.where(detect_anchors == sort_anchors[0])[0])
        anchor_2_temp = int(np.where(detect_anchors == sort_anchors[1])[0])

    if anchor_1_temp > anchor_2_temp:
        anchor_1 = anchor_1_temp
        anchor_2 = anchor_2_temp

    else:
        anchor_1 = anchor_2_temp
        anchor_2 = anchor_1_temp

    X = np.delete(X, anchor_1, 1)
    X = np.delete(X, anchor_2, 1)

    X = X.T

    if savgol_filter:
        X_med = scipy.signal.savgol_filter(X, cfg['savgol_length'], cfg['savgol_order'])
    else:
        X_med = X

    num_frames = len(X_med.T)
    test = int(num_frames*testfraction)

    z_test =X_med[:,:test]
    z_train = X_med[:,test:]

    if check_parameter:
        plot_check_parameter(cfg, iqr_val, num_frames, X_true, X_med)

    else:
        #save numpy arrays the the test/train info:
        np.save(os.path.join(cfg['project_path'],"data", "train",'train_seq.npy'), z_train)
        np.save(os.path.join(cfg['project_path'],"data", "train", 'test_seq.npy'), z_test)

        for i, file in enumerate(files):
            np.save(os.path.join(cfg['project_path'],"data", file, file+'-PE-seq-clean.npy'), X_med[:,pos[i]:pos[i+1]])

        logger.info('Lenght of train data: %d' %len(z_train.T))
        logger.info('Lenght of test data: %d' %len(z_test.T))


def traindata_fixed(
    cfg: dict,
    files: List[str],
    testfraction: float,
    num_features: int,
    savgol_filter: bool,
    check_parameter: bool,
    pose_ref_index: Optional[List[int]]
) -> None:
    """
    Create training dataset for fixed data.

    Args:
        cfg (dict): Configuration parameters.
        files (List[str]): List of files.
        testfraction (float): Fraction of data to use as test data.
        num_features (int): Number of features.
        savgol_filter (bool): Flag indicating whether to apply Savitzky-Golay filter.
        check_parameter (bool): If True, the function will plot the z-scored data and the filtered data.
        pose_ref_index (Optional[List[int]]): List of reference coordinate indices for alignment.

    Returns:
        None - Save numpy arrays with the test/train info to the project folder.
    """
    X_train = []
    pos = []
    pos_temp = 0
    pos.append(0)

    if check_parameter:
        X_true = []
        rnd_file = np.random.choice(len(files))
        files = [files[0]]

    for file in files:
        logger.info("z-scoring of file %s" %file)
        path_to_file = os.path.join(cfg['project_path'],"data", file, file+'-PE-seq.npy')
        data = np.load(path_to_file)
        X_mean = np.mean(data,axis=None)
        X_std = np.std(data, axis=None)
        X_z = (data.T - X_mean) / X_std

        if check_parameter:
            X_z_copy = X_z.copy()
            X_true.append(X_z_copy)

        if cfg['robust']:
            iqr_val = iqr(X_z)
            logger.info("IQR value: %.2f, IQR cutoff: %.2f" %(iqr_val, cfg['iqr_factor']*iqr_val))
            for i in range(X_z.shape[0]):
                for marker in range(X_z.shape[1]):
                    if X_z[i,marker] > cfg['iqr_factor']*iqr_val:
                        X_z[i,marker] = np.nan

                    elif X_z[i,marker] < -cfg['iqr_factor']*iqr_val:
                        X_z[i,marker] = np.nan

                X_z[i,:] = interpol_all_nans(X_z[i,:])

        X_len = len(data.T)
        pos_temp += X_len
        pos.append(pos_temp)
        X_train.append(X_z)

    X = np.concatenate(X_train, axis=0).T

    if savgol_filter:
        X_med = scipy.signal.savgol_filter(X, cfg['savgol_length'], cfg['savgol_order'])
    else:
        X_med = X

    num_frames = len(X_med.T)
    test = int(num_frames*testfraction)

    z_test = X_med[:,:test]
    z_train = X_med[:,test:]

    if check_parameter:
        plot_check_parameter(cfg, iqr_val, num_frames, X_true, X_med)

    else:
        if pose_ref_index is None:
            raise ValueError("Please provide a pose reference index for training on fixed data. E.g. [0,5]")
        #save numpy arrays the the test/train info:
        np.save(os.path.join(cfg['project_path'],"data", "train",'train_seq.npy'), z_train)
        np.save(os.path.join(cfg['project_path'],"data", "train", 'test_seq.npy'), z_test)

        y_shifted_indices = np.arange(0, num_features, 2)
        x_shifted_indices = np.arange(1, num_features, 2)
        belly_Y_ind = pose_ref_index[0] * 2
        belly_X_ind = (pose_ref_index[0] * 2) + 1

        for i, file in enumerate(files):

            # Shifting section added 2/29/2024 PN
            X_med_shifted_file = X_med[:, pos[i]:pos[i+1]]
            belly_Y_shift = X_med[belly_Y_ind, pos[i]:pos[i+1]]
            belly_X_shift = X_med[belly_X_ind, pos[i]:pos[i+1]]

            X_med_shifted_file[y_shifted_indices,:] -= belly_Y_shift
            X_med_shifted_file[x_shifted_indices,:] -= belly_X_shift

            np.save(os.path.join(cfg['project_path'],"data", file, file+'-PE-seq-clean.npy'), X_med_shifted_file) # saving new shifted file

        logger.info('Lenght of train data: %d' %len(z_train.T))
        logger.info('Lenght of test data: %d' %len(z_test.T))


@save_state(model=CreateTrainsetFunctionSchema)
def create_trainset(
    config: str,
    pose_ref_index: Optional[List] = None,
    check_parameter: bool =False,
    save_logs: bool = False
) -> None:
    """Creates a training dataset for the VAME model.

    Args:
        config (str): Path to the config file.
        pose_ref_index (Optional[List], optional): List of reference coordinate indices for alignment. Defaults to None.
        check_parameter (bool, optional): If True, the function will plot the z-scored data and the filtered data. Defaults to False.
    """

    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        fixed = cfg['egocentric_data']

        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'create_trainset.log'
            logger_config.add_file_handler(log_path)


        if not os.path.exists(os.path.join(cfg['project_path'],'data','train',"")):
            os.mkdir(os.path.join(cfg['project_path'],'data','train',""))

        files = []
        if cfg['all_data'] == 'No':
            for file in cfg['video_sets']:
                use_file = input("Do you want to train on " + file + "? yes/no: ")
                if use_file == 'yes':
                    files.append(file)
                if use_file == 'no':
                    continue
        else:
            for file in cfg['video_sets']:
                files.append(file)

        logger.info("Creating training dataset...")
        if cfg['robust']:
            logger.info("Using robust setting to eliminate outliers! IQR factor: %d" %cfg['iqr_factor'])

        if not fixed:
            logger.info("Creating trainset from the vame.egocentrical_alignment() output ")
            traindata_aligned(cfg, files, cfg['test_fraction'], cfg['savgol_filter'], check_parameter)
        else:
            logger.info("Creating trainset from the vame.csv_to_numpy() output ")
            traindata_fixed(cfg, files, cfg['test_fraction'], cfg['num_features'], cfg['savgol_filter'], check_parameter,  pose_ref_index)

        if not check_parameter:
            logger.info("A training and test set has been created. Next step: vame.train_model()")

    except Exception as e:
        logger.exception(str(e))
        raise e
    finally:
        logger_config.remove_file_handler()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import umap
import numpy as np
from pathlib import Path
import matplotlib.pyplot as plt
from typing import Optional, Union, Dict
from vame.util.auxiliary import read_config
from vame.schemas.states import VisualizationFunctionSchema, save_state
from vame.logging.logger import VameLogger
from vame.schemas.project import Parametrizations


logger_config = VameLogger(__name__)
logger = logger_config.logger


def umap_embedding(cfg: dict, file: str, model_name: str, n_cluster: int, parametrization: str) -> np.ndarray:
    """Perform UMAP embedding for given file and parameters.

    Args:
        cfg (dict): Configuration parameters.
        file (str): File path.
        model_name (str): Model name.
        n_cluster (int): Number of clusters.
        parametrization (str): parametrization.

    Returns:
        np.ndarray: UMAP embedding.
    """
    reducer = umap.UMAP(
        n_components=2,
        min_dist=cfg['min_dist'],
        n_neighbors=cfg['n_neighbors'],
        random_state=cfg['random_state']
    )

    logger.info("UMAP calculation for file %s" %file)

    folder = os.path.join(cfg['project_path'],"results",file,model_name, parametrization +'-'+str(n_cluster),"")
    latent_vector = np.load(os.path.join(folder,'latent_vector_'+file+'.npy'))

    num_points = cfg['num_points']
    if num_points > latent_vector.shape[0]:
        num_points = latent_vector.shape[0]
    logger.info("Embedding %d data points.." %num_points)

    embed = reducer.fit_transform(latent_vector[:num_points,:])
    np.save(os.path.join(folder,"community","umap_embedding_"+file+'.npy'), embed)

    return embed


# def umap_vis_community_labels(cfg: dict, embed: np.ndarray, community_labels_all: np.ndarray, save_path: str | None) -> None:
#     """Create plotly visualizaton of UMAP embedding with community labels.

#     Args:
#         cfg (dict): Configuration parameters.
#         embed (np.ndarray): UMAP embedding.
#         community_labels_all (np.ndarray): Community labels.
#         save_path: Path to save the plot. If None it will not save the plot.

#     Returns:
#         None
#     """
#     num_points = cfg['num_points']
#     community_labels_all = np.asarray(community_labels_all)
#     if num_points > community_labels_all.shape[0]:
#         num_points = community_labels_all.shape[0]
#     logger.info("Embedding %d data points.." %num_points)

#     num = np.unique(community_labels_all)

#     fig = plt.figure(1)
#     plt.scatter(
#         embed[:,0],
#         embed[:,1],
#         c=community_labels_all[:num_points],
#         cmap='Spectral',
#         s=2,
#         alpha=1
#     )
#     plt.colorbar(boundaries=np.arange(np.max(num)+2)-0.5).set_ticks(np.arange(np.max(num)+1))
#     plt.gca().set_aspect('equal', 'datalim')
#     plt.grid(False)

#     if save_path is not None:
#         plt.savefig(save_path)
#         return fig
#     plt.show()
#     return fig


def umap_vis(embed: np.ndarray, num_points: int) -> None:
    """
    Visualize UMAP embedding without labels.

    Args:
        embed (np.ndarray): UMAP embedding.
        num_points (int): Number of data points to visualize.

    Returns:
        None - Plot Visualization of UMAP embedding.
    """
    #plt.cla()
    #plt.clf()
    plt.close('all')
    fig = plt.figure(1)
    plt.scatter(embed[:num_points,0], embed[:num_points,1], s=2, alpha=.5)
    plt.gca().set_aspect('equal', 'datalim')
    plt.grid(False)
    return fig


def umap_label_vis(embed: np.ndarray, label: np.ndarray, n_cluster: int, num_points: int) -> None:
    """
    Visualize UMAP embedding with motif labels.

    Args:
        embed (np.ndarray): UMAP embedding.
        label (np.ndarray): Motif labels.
        n_cluster (int): Number of clusters.
        num_points (int): Number of data points to visualize.

    Returns:
        fig - Plot figure of UMAP visualization embedding with motif labels.
    """
    fig = plt.figure(1)
    plt.scatter(embed[:num_points,0], embed[:num_points,1],  c=label[:num_points], cmap='Spectral', s=2, alpha=.7)
    #plt.colorbar(boundaries=np.arange(n_cluster+1)-0.5).set_ticks(np.arange(n_cluster))
    plt.gca().set_aspect('equal', 'datalim')
    plt.grid(False)
    return fig


def umap_vis_comm(embed: np.ndarray, community_label: np.ndarray, num_points: int) -> None:
    """
    Visualize UMAP embedding with community labels.

    Args:
        embed (np.ndarray): UMAP embedding.
        community_label (np.ndarray): Community labels.
        num_points (int): Number of data points to visualize.

    Returns:
        fig - Plot figure of UMAP visualization embedding with community labels.
    """
    num = np.unique(community_label).shape[0]
    fig = plt.figure(1)
    plt.scatter(embed[:num_points,0], embed[:num_points,1],  c=community_label[:num_points], cmap='Spectral', s=2, alpha=.7)
    #plt.colorbar(boundaries=np.arange(num+1)-0.5).set_ticks(np.arange(num))
    plt.gca().set_aspect('equal', 'datalim')
    plt.grid(False)
    return fig

@save_state(model=VisualizationFunctionSchema)
def visualization(
    config: Union[str, Path],
    parametrization: Parametrizations,
    label: Optional[str] = None,
    save_logs: bool = False
) -> None:
    """
    Visualize UMAP embeddings based on configuration settings.

    Args:
        config (Union[str, Path]): Path to the configuration file.
        label (str, optional): Type of labels to visualize. Default is None.

    Returns:
        None - Plot Visualization of UMAP embeddings.
    """
    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        parametrizations = cfg['parametrizations']

        if save_logs:
            logs_path = Path(cfg['project_path']) / "logs" / 'visualization.log'
            logger_config.add_file_handler(logs_path)

        model_name = cfg['model_name']
        n_cluster = cfg['n_cluster']

        files = []
        if cfg['all_data'] == 'No':
            all_flag = input("Do you want to write motif videos for your entire dataset? \n"
                        "If you only want to use a specific dataset type filename: \n"
                        "yes/no/filename ")
        else:
            all_flag = 'yes'

        if all_flag == 'yes' or all_flag == 'Yes':
            for file in cfg['video_sets']:
                files.append(file)

        elif all_flag == 'no' or all_flag == 'No':
            for file in cfg['video_sets']:
                use_file = input("Do you want to quantify " + file + "? yes/no: ")
                if use_file == 'yes':
                    files.append(file)
                if use_file == 'no':
                    continue
        else:
            files.append(all_flag)


        for idx, file in enumerate(files):
            path_to_file=os.path.join(cfg['project_path'],"results",file,"",model_name,"",parametrization+'-'+str(n_cluster))

            try:
                embed = np.load(os.path.join(path_to_file,"","community","","umap_embedding_"+file+".npy"))
                num_points = cfg['num_points']
                if num_points > embed.shape[0]:
                    num_points = embed.shape[0]
            except Exception:
                if not os.path.exists(os.path.join(path_to_file,"community")):
                    os.mkdir(os.path.join(path_to_file,"community"))
                logger.info("Compute embedding for file %s" %file)
                embed = umap_embedding(cfg, file, model_name, n_cluster, parametrization)
                num_points = cfg['num_points']
                if num_points > embed.shape[0]:
                    num_points = embed.shape[0]

            if label is None:
                output_figure = umap_vis(embed, num_points)
                fig_path = os.path.join(path_to_file,"community","umap_vis_label_none_"+file+".png")
                output_figure.savefig(fig_path)

            if label == 'motif':
                motif_label = np.load(os.path.join(path_to_file,"",str(n_cluster)+'_' + parametrization + '_label_'+file+'.npy'))
                output_figure = umap_label_vis(embed, motif_label, n_cluster, num_points)
                fig_path = os.path.join(path_to_file,"community","umap_vis_motif_"+file+".png")
                output_figure.savefig(fig_path)

            if label == "community":
                community_label = np.load(os.path.join(path_to_file,"","community","","community_label_"+file+".npy"))
                output_figure = umap_vis_comm(embed, community_label, num_points)
                fig_path = os.path.join(path_to_file,"community","umap_vis_community_"+file+".png")
                output_figure.savefig(fig_path)
    except Exception as e:
        logger.exception(str(e))
        raise e
    finally:
        logger_config.remove_file_handler()














#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import tqdm
import torch
import pickle
import numpy as np
from pathlib import Path
from typing import List, Tuple
from vame.util.data_manipulation import consecutive
from hmmlearn import hmm
from sklearn.cluster import KMeans
from vame.schemas.states import save_state, PoseSegmentationFunctionSchema
from vame.logging.logger import VameLogger, TqdmToLogger
from vame.util.auxiliary import read_config
from vame.model.rnn_model import RNN_VAE
from vame.util.model_util import load_model


logger_config = VameLogger(__name__)
logger = logger_config.logger


def embedd_latent_vectors(cfg: dict, files: List[str], model: RNN_VAE, fixed: bool, tqdm_stream: TqdmToLogger | None) -> List[np.ndarray]:
    """Embed latent vectors for the given files using the VAME model.

    Args:
        cfg (dict): Configuration dictionary.
        files (List[str]): List of files names.
        model (RNN_VAE): VAME model.
        fixed (bool): Whether the model is fixed.
        tqdm_stream (TqdmToLogger): TQDM Stream to redirect the tqdm output to logger.

    Returns:
        List[np.ndarray]: List of latent vectors for each file.
    """
    project_path = cfg['project_path']
    temp_win = cfg['time_window']
    num_features = cfg['num_features']
    if not fixed:
        num_features = num_features - 2

    use_gpu = torch.cuda.is_available()
    if use_gpu:
        pass
    else:
        torch.device("cpu")

    latent_vector_files = []

    for file in files:
        logger.info('Embedding of latent vector for file %s' %file)
        data = np.load(os.path.join(project_path,'data',file,file+'-PE-seq-clean.npy'))
        latent_vector_list = []
        with torch.no_grad():
            for i in tqdm.tqdm(range(data.shape[1] - temp_win), file=tqdm_stream):
            # for i in tqdm.tqdm(range(10000)):
                data_sample_np = data[:,i:temp_win+i].T
                data_sample_np = np.reshape(data_sample_np, (1, temp_win, num_features))
                if use_gpu:
                    h_n = model.encoder(torch.from_numpy(data_sample_np).type('torch.FloatTensor').cuda())
                else:
                    h_n = model.encoder(torch.from_numpy(data_sample_np).type('torch.FloatTensor').to())
                mu, _, _ = model.lmbda(h_n)
                latent_vector_list.append(mu.cpu().data.numpy())

        latent_vector = np.concatenate(latent_vector_list, axis=0)
        latent_vector_files.append(latent_vector)

    return latent_vector_files


def get_motif_usage(label: np.ndarray) -> np.ndarray:
    """Compute motif usage from the label array.

    Args:
        label (np.ndarray): Label array.

    Returns:
        np.ndarray: Array of motif usage counts.
    """
    motif_usage = np.unique(label, return_counts=True)
    cons = consecutive(motif_usage[0])
    if len(cons) != 1:
        usage_list = list(motif_usage[1])
        for i in range(len(cons)-1):
            a = cons[i+1][0]
            b = cons[i][-1]
            d = (a-b)-1
            for j in range(1,d+1):
                index = cons[i][-1]+j
                usage_list.insert(index,0)
        usage = np.array(usage_list)
        motif_usage = usage
    else:
        motif_usage = motif_usage[1]

    return motif_usage


def same_parametrization(
    cfg: dict,
    files: List[str],
    latent_vector_files: List[np.ndarray],
    states: int,
    parametrization: str
) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    """Apply the same parametrization to all animals.

    Args:
        cfg (dict): Configuration dictionary.
        files (List[str]): List of file names.
        latent_vector_files (List[np.ndarray]): List of latent vector arrays.
        states (int): Number of states.
        parametrization (str): parametrization method.

    Returns:
        Tuple: Tuple of labels, cluster centers, and motif usages.
    """
    labels = []
    cluster_centers = []
    motif_usages = []

    latent_vector_cat = np.concatenate(latent_vector_files, axis=0)
    if parametrization == "kmeans":
        logger.info("Using kmeans as parametrization!")
        kmeans = KMeans(init='k-means++', n_clusters=states, random_state=42, n_init=20).fit(latent_vector_cat)
        clust_center = kmeans.cluster_centers_
        label = kmeans.predict(latent_vector_cat)

    elif parametrization == "hmm":
        if not cfg['hmm_trained']:
            logger.info("Using a HMM as parametrization!")
            hmm_model = hmm.GaussianHMM(n_components=states, covariance_type="full", n_iter=100)
            hmm_model.fit(latent_vector_cat)
            label = hmm_model.predict(latent_vector_cat)
            save_data = os.path.join(cfg['project_path'], "results", "")
            with open(save_data+"hmm_trained.pkl", "wb") as file:
                pickle.dump(hmm_model, file)
        else:
            logger.info("Using a pretrained HMM as parametrization!")
            save_data = os.path.join(cfg['project_path'], "results", "")
            with open(save_data+"hmm_trained.pkl", "rb") as file:
                hmm_model = pickle.load(file)
            label = hmm_model.predict(latent_vector_cat)

    idx = 0
    for i, file in enumerate(files):
        file_len = latent_vector_files[i].shape[0]
        labels.append(label[idx:idx+file_len])
        if parametrization == "kmeans":
            cluster_centers.append(clust_center)

        motif_usage = get_motif_usage(label[idx:idx+file_len])
        motif_usages.append(motif_usage)
        idx += file_len

    return labels, cluster_centers, motif_usages


def individual_parametrization(
    cfg: dict,
    files: List[str],
    latent_vector_files: List[np.ndarray],
    cluster: int,
) -> Tuple:
    """Apply individual parametrization to each animal.

    Args:
        cfg (dict): Configuration dictionary.
        files (List[str]): List of file names.
        latent_vector_files (List[np.ndarray]): List of latent vector arrays.
        cluster (int): Number of clusters.

    Returns:
        Tuple: Tuple of labels, cluster centers, and motif usages.
    """
    random_state = cfg['random_state_kmeans']
    n_init = cfg['n_init_kmeans']

    labels = []
    cluster_centers = []
    motif_usages = []
    for i, file in enumerate(files):
        logger.info(f'Processing file: {file}')
        kmeans = KMeans(init='k-means++', n_clusters=cluster, random_state=random_state, n_init=n_init).fit(latent_vector_files[i])
        clust_center = kmeans.cluster_centers_
        label = kmeans.predict(latent_vector_files[i])
        motif_usage = get_motif_usage(label)
        motif_usages.append(motif_usage)
        labels.append(label)
        cluster_centers.append(clust_center)

    return labels, cluster_centers, motif_usages

@save_state(model=PoseSegmentationFunctionSchema)
def pose_segmentation(config: str, save_logs: bool = False) -> None:
    """Perform pose segmentation using the VAME model.

    Args:
        config (str): Path to the configuration file.

    Returns:
        None
    """
    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        tqdm_stream = None
        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'pose_segmentation.log'
            logger_config.add_file_handler(log_path)
            tqdm_stream = TqdmToLogger(logger)
        model_name = cfg['model_name']
        n_cluster = cfg['n_cluster']
        fixed = cfg['egocentric_data']
        parametrizations = cfg['parametrizations']

        logger.info('Pose segmentation for VAME model: %s \n' %model_name)
        ind_param = cfg['individual_parametrization']

        logger.info(f"parametrizations: {parametrizations}")

        for parametrization in parametrizations:
            logger.info(f'Running pose segmentation using {parametrization} parametrization')
            for folders in cfg['video_sets']:
                if not os.path.exists(os.path.join(cfg['project_path'],"results",folders,model_name,"")):
                    os.mkdir(os.path.join(cfg['project_path'],"results",folders,model_name,""))

            files = []
            if cfg['all_data'] == 'No':
                all_flag = input("Do you want to qunatify your entire dataset? \n"
                                "If you only want to use a specific dataset type filename: \n"
                                "yes/no/filename ")
                file = all_flag

            else:
                all_flag = 'yes'

            if all_flag == 'yes' or all_flag == 'Yes':
                for file in cfg['video_sets']:
                    files.append(file)
            elif all_flag == 'no' or all_flag == 'No':
                for file in cfg['video_sets']:
                    use_file = input("Do you want to quantify " + file + "? yes/no: ")
                    if use_file == 'yes':
                        files.append(file)
                    if use_file == 'no':
                        continue
            else:
                files.append(all_flag)
            # files.append("mouse-3-1")
            # file="mouse-3-1"

            use_gpu = torch.cuda.is_available()
            if use_gpu:
                logger.info("Using CUDA")
                logger.info('GPU active: {}'.format(torch.cuda.is_available()))
                logger.info('GPU used: {}'.format(torch.cuda.get_device_name(0)))
            else:
                logger.info("CUDA is not working! Attempting to use the CPU...")
                torch.device("cpu")

            if not os.path.exists(os.path.join(cfg['project_path'],"results",file,model_name, parametrization+'-'+str(n_cluster),"")):
                new = True
                model = load_model(cfg, model_name, fixed)
                latent_vectors = embedd_latent_vectors(cfg, files, model, fixed, tqdm_stream=tqdm_stream)

                if not ind_param:
                    logger.info("For all animals the same parametrization of latent vectors is applied for %d cluster" %n_cluster)
                    labels, cluster_center, motif_usages = same_parametrization(cfg, files, latent_vectors, n_cluster, parametrization)
                else:
                    logger.info("Individual parametrization of latent vectors for %d cluster" %n_cluster)
                    labels, cluster_center, motif_usages = individual_parametrization(cfg, files, latent_vectors, n_cluster)

            else:
                logger.info('\n'
                    'For model %s a latent vector embedding already exists. \n'
                    'parametrization of latent vector with %d k-Means cluster' %(model_name, n_cluster))

                if os.path.exists(os.path.join(cfg['project_path'],"results",file,model_name, parametrization+'-'+str(n_cluster),"")):
                    flag = input('WARNING: A parametrization for the chosen cluster size of the model already exists! \n'
                                'Do you want to continue? A new parametrization will be computed! (yes/no) ')
                else:
                    flag = 'yes'

                if flag == 'yes':
                    new = True
                    latent_vectors = []
                    for file in files:
                        path_to_latent_vector = os.path.join(cfg['project_path'],"results",file,model_name, parametrization+'-'+str(n_cluster),"")
                        latent_vector = np.load(os.path.join(path_to_latent_vector,'latent_vector_'+file+'.npy'))
                        latent_vectors.append(latent_vector)

                    if not ind_param:
                        logger.info("For all animals the same parametrization of latent vectors is applied for %d cluster" %n_cluster)
                        labels, cluster_center, motif_usages = same_parametrization(cfg, files, latent_vectors, n_cluster, parametrization)
                    else:
                        logger.info("Individual parametrization of latent vectors for %d cluster" %n_cluster)
                        labels, cluster_center, motif_usages = individual_parametrization(cfg, files, latent_vectors, n_cluster)

                else:
                    logger.info('No new parametrization has been calculated.')
                    new = False

            if new:
                for idx, file in enumerate(files):
                    logger.info(os.path.join(cfg['project_path'],"results",file,"",model_name,parametrization+'-'+str(n_cluster),""))
                    if not os.path.exists(os.path.join(cfg['project_path'],"results",file,model_name,parametrization+'-'+str(n_cluster),"")):
                        try:
                            os.mkdir(os.path.join(cfg['project_path'],"results",file,"",model_name,parametrization+'-'+str(n_cluster),""))
                        except OSError as error:
                            logger.error(error)

                    save_data = os.path.join(cfg['project_path'],"results",file,model_name,parametrization+'-'+str(n_cluster),"")
                    np.save(os.path.join(save_data,str(n_cluster)+'_' + parametrization + '_label_'+file), labels[idx])
                    if parametrization=="kmeans":
                        np.save(os.path.join(save_data,'cluster_center_'+file), cluster_center[idx])
                    np.save(os.path.join(save_data,'latent_vector_'+file), latent_vectors[idx])
                    np.save(os.path.join(save_data,'motif_usage_'+file), motif_usages[idx])


                logger.info("You succesfully extracted motifs with VAME! From here, you can proceed running vame.motif_videos() ")
                    # "to get an idea of the behavior captured by VAME. This will leave you with short snippets of certain movements."
                    # "To get the full picture of the spatiotemporal dynamic we recommend applying our community approach afterwards.")
    except Exception as e:
        logger.exception(f"An error occurred during pose segmentation: {e}")
    finally:
        logger_config.remove_file_handler()




from pydantic import BaseModel, Field, ConfigDict
from typing import List
from enum import Enum

class Parametrizations(str, Enum):
    hmm = 'hmm'
    kmeans = 'kmeans'

    class Config:
        use_enum_values = True

class ProjectSchema(BaseModel):
    # Project attributes
    Project: str = Field(..., title='Project name')
    model_name: str = Field(default='VAME', title='Model name')
    n_cluster: int = Field(default=15, title='Number of clusters')
    pose_confidence: float = Field(default=0.99, title='Pose confidence')

    # Project path and videos
    project_path: str = Field(..., title='Project path')
    video_sets: List[str] = Field(..., title='Video sets')

    # Data
    all_data: str = Field(default='yes', title='All data')
    egocentric_data: bool = Field(default=False, title='Egocentric data')
    robust: bool = Field(default=True, title='Robust data')
    iqr_factor: int = Field(default=4, title='IQR factor')
    axis: str = Field(default='None', title='Axis')
    savgol_filter: bool = Field(default=True, title='Savgol filter')
    savgol_length: int = Field(default=5, title='Savgol length')
    savgol_order: int = Field(default=2, title='Savgol order')
    test_fraction: float = Field(default=0.1, title='Test fraction')

    # RNN model general hyperparameter:
    pretrained_model: str = Field(default='None', title='Pretrained model')
    pretrained_weights: bool = Field(default=False, title='Pretrained weights')
    num_features: int = Field(default=12, title='Number of features')
    batch_size: int = Field(default=256, title='Batch size')
    max_epochs: int = Field(default=500, title='Max epochs')
    model_snapshot: int = Field(default=50, title='Model snapshot')
    model_convergence: int = Field(default=50, title='Model convergence')
    transition_function: str = Field(default='GRU', title='Transition function')
    beta: float = Field(default=1, title='Beta')
    beta_norm: bool = Field(default=False, title='Beta normalization')
    zdims: int = Field(default=30, title='Zdims')
    learning_rate: float = Field(default=5e-4, title='Learning rate')
    time_window: int = Field(default=30, title='Time window')
    prediction_decoder: int = Field(default=1, title='Prediction decoder')
    prediction_steps: int = Field(default=15, title='Prediction steps')
    noise: bool = Field(default=False, title='Noise')
    scheduler: int = Field(default=1, title='Scheduler')
    scheduler_step_size: int = Field(default=100, title='Scheduler step size')
    scheduler_gamma: float = Field(default=0.2, title='Scheduler gamma')
    scheduler_threshold: float = Field(default=None, title='Scheduler threshold')
    softplus: bool = Field(default=False, title='Softplus')

    # Segmentation:
    parametrizations: List[Parametrizations] = Field(
        title='Parametrizations',
        default_factory=lambda: [Parametrizations.hmm.value, Parametrizations.kmeans.value]
    )
    hmm_trained: bool = Field(default=False, title='HMM trained')
    load_data: str = Field(default='-PE-seq-clean', title='Load data')
    individual_parametrization: bool = Field(default=False, title='Individual parametrization')
    random_state_kmeans: int = Field(default=42, title='Random state kmeans')
    n_init_kmeans: int = Field(default=15, title='N init kmeans')

    # Video writer:
    length_of_motif_video: int = Field(default=1000, title='Length of motif video')

    # UMAP parameter:
    min_dist: float = Field(default=0.1, title='Min dist')
    n_neighbors: int = Field(default=200, title='N neighbors')
    random_state: int = Field(default=42, title='Random state')
    num_points: int = Field(default=30000, title='Num points')

    # RNN encoder hyperparameter:
    hidden_size_layer_1: int = Field(default=256, title='Hidden size layer 1')
    hidden_size_layer_2: int = Field(default=256, title='Hidden size layer 2')
    dropout_encoder: float = Field(default=0, title='Dropout encoder')

    # RNN reconstruction hyperparameter:
    hidden_size_rec: int = Field(default=256, title='Hidden size rec')
    dropout_rec: float = Field(default=0, title='Dropout rec')
    n_layers: int = Field(default=1, title='N layers')

    # RNN prediction hyperparameter:
    hidden_size_pred: int = Field(default=256, title='Hidden size pred')
    dropout_pred: float = Field(default=0, title='Dropout pred')

    # RNN loss hyperparameter:
    mse_reconstruction_reduction: str = Field(default='sum', title='MSE reconstruction reduction')
    mse_prediction_reduction: str = Field(default='sum', title='MSE prediction reduction')
    kmeans_loss: int = Field(default=30, title='Kmeans loss')
    kmeans_lambda: float = Field(default=0.1, title='Kmeans lambda')
    anneal_function: str = Field(default='linear', title='Anneal function')
    kl_start: int = Field(default=2, title='KL start')
    annealtime: int = Field(default=4, title='Annealtime')

    model_config: ConfigDict = ConfigDict(protected_namespaces=())


import numpy as np
from typing import List, Tuple
import cv2 as cv
import os
from scipy.ndimage import median_filter
import tqdm
from vame.logging.logger import VameLogger


logger_config = VameLogger(__name__)
logger = logger_config.logger

def consecutive(data: np.ndarray, stepsize: int = 1) -> List[np.ndarray]:
    """Find consecutive sequences in the data array.

    Args:
        data (np.ndarray): Input array.
        stepsize (int, optional): Step size. Defaults to 1.

    Returns:
        List[np.ndarray]: List of consecutive sequences.
    """
    data = data[:]
    return np.split(data, np.where(np.diff(data) != stepsize)[0]+1)


def nan_helper(y: np.ndarray) -> Tuple:
    """
    Identifies indices of NaN values in an array and provides a function to convert them to non-NaN indices.

    Args:
        y (np.ndarray): Input array containing NaN values.

    Returns:
        Tuple[np.ndarray, Union[np.ndarray, None]]: A tuple containing two elements:
            - An array of boolean values indicating the positions of NaN values.
            - A lambda function to convert NaN indices to non-NaN indices.
    """
    return np.isnan(y), lambda z: z.nonzero()[0]


def interpol_all_nans(arr: np.ndarray) -> np.ndarray:
    """
    Interpolates all NaN values in the given array.

    Args:
        arr (np.ndarray): Input array containing NaN values.

    Returns:
        np.ndarray: Array with NaN values replaced by interpolated values.
    """
    y = np.transpose(arr)
    nans, x = nan_helper(y)
    y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    arr = np.transpose(y)
    return arr


def interpol_first_rows_nans(arr: np.ndarray) -> np.ndarray:
    """
    Interpolates NaN values in the given array.

    Args:
        arr (np.ndarray): Input array with NaN values.

    Returns:
        np.ndarray: Array with interpolated NaN values.
    """

    y = np.transpose(arr)

    nans, x = nan_helper(y[0])
    y[0][nans]= np.interp(x(nans), x(~nans), y[0][~nans])
    nans, x = nan_helper(y[1])
    y[1][nans]= np.interp(x(nans), x(~nans), y[1][~nans])

    arr = np.transpose(y)

    return arr

def crop_and_flip(
    rect: Tuple,
    src: np.ndarray,
    points: List[np.ndarray],
    ref_index: Tuple[int, int]
) -> Tuple[np.ndarray, List[np.ndarray]]:
    """
    Crop and flip the image based on the given rectangle and points.

    Args:
        rect (Tuple): Rectangle coordinates (center, size, theta).
        src (np.ndarray): Source image.
        points (List[np.ndarray]): List of points.
        ref_index (Tuple[int, int]): Reference indices for alignment.

    Returns:
        Tuple[np.ndarray, List[np.ndarray]]: Cropped and flipped image, and shifted points.
    """
    #Read out rect structures and convert
    center, size, theta = rect

    center, size = tuple(map(int, center)), tuple(map(int, size))

    #Get rotation matrix
    M = cv.getRotationMatrix2D(center, theta, 1)

    #shift DLC points
    x_diff = center[0] - size[0]//2
    y_diff = center[1] - size[1]//2

    dlc_points_shifted = []

    for i in points:
        point=cv.transform(np.array([[[i[0], i[1]]]]),M)[0][0]

        point[0] -= x_diff
        point[1] -= y_diff

        dlc_points_shifted.append(point)

    # Perform rotation on src image
    dst = cv.warpAffine(src.astype('float32'), M, src.shape[:2])
    out = cv.getRectSubPix(dst, size, center)

    #check if flipped correctly, otherwise flip again
    if dlc_points_shifted[ref_index[1]][0] >= dlc_points_shifted[ref_index[0]][0]:
        rect = ((size[0]//2,size[0]//2),size,180) #should second value be size[1]? Is this relevant to the flip? 3/5/24 KKL
        center, size, theta = rect
        center, size = tuple(map(int, center)), tuple(map(int, size))

        #Get rotation matrix
        M = cv.getRotationMatrix2D(center, theta, 1)

        #shift DLC points
        x_diff = center[0] - size[0]//2
        y_diff = center[1] - size[1]//2

        points = dlc_points_shifted
        dlc_points_shifted = []

        for i in points:
            point=cv.transform(np.array([[[i[0], i[1]]]]),M)[0][0]

            point[0] -= x_diff
            point[1] -= y_diff

            dlc_points_shifted.append(point)

        # Perform rotation on src image
        dst = cv.warpAffine(out.astype('float32'), M, out.shape[:2])
        out = cv.getRectSubPix(dst, size, center)

    return out, dlc_points_shifted

def background(
    path_to_file: str,
    filename: str,
    file_format: str = '.mp4',
    num_frames: int = 1000,
    save_background: bool = True
) -> np.ndarray:
    """
    Compute background image from fixed camera.

    Args:
        path_to_file (str): Path to the directory containing the video files.
        filename (str): Name of the video file.
        file_format (str, optional): Format of the video file. Defaults to '.mp4'.
        num_frames (int, optional): Number of frames to use for background computation. Defaults to 1000.

    Returns:
        np.ndarray: Background image.
    """

    capture = cv.VideoCapture(os.path.join(path_to_file,"videos",filename+file_format))

    if not capture.isOpened():
        raise Exception("Unable to open video file: {0}".format(os.path.join(path_to_file,"videos",filename+file_format)))

    frame_count = int(capture.get(cv.CAP_PROP_FRAME_COUNT))
    ret, frame = capture.read()

    height, width, _ = frame.shape
    frames = np.zeros((height,width,num_frames))

    for i in tqdm.tqdm(range(num_frames), disable=not True, desc='Compute background image for video %s' %filename):
        rand = np.random.choice(frame_count, replace=False)
        capture.set(1,rand)
        ret, frame = capture.read()
        gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
        frames[...,i] = gray

    logger.info('Finishing up!')
    medFrame = np.median(frames,2)
    background = median_filter(medFrame, (5,5))

    if save_background:
        np.save(os.path.join(path_to_file,"videos",filename+'-background.npy'),background)

    capture.release()
    return background


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0

The following code is adapted from:

DeepLabCut2.0 Toolbox (deeplabcut.org)
© A. & M. Mathis Labs
https://github.com/AlexEMG/DeepLabCut
Please see AUTHORS for contributors.
https://github.com/AlexEMG/DeepLabCut/blob/master/AUTHORS
Licensed under GNU Lesser General Public License v3.0
"""

import os
from pathlib import Path
import shutil
from datetime import datetime as dt
from vame.util.auxiliary import write_config
from typing import List
from vame.schemas.project import ProjectSchema
from vame.schemas.states import VAMEPipelineStatesSchema
import json
from vame.logging.logger import VameLogger

logger_config = VameLogger(__name__)
logger = logger_config.logger



def init_new_project(
    project: str,
    videos: List[str],
    poses_estimations: List[str],
    working_directory: str = '.',
    videotype: str = '.mp4'
) -> str:
    """Creates a new VAME project with the given parameters.

    Args:
        project (str): Project name.
        videos (List[str]): List of videos paths to be used in the project. E.g. ['./sample_data/Session001.mp4']
        poses_estimations (List[str]): List of pose estimation files paths to be used in the project. E.g. ['./sample_data/pose estimation/Session001.csv']        working_directory (str, optional): _description_. Defaults to None.
        videotype (str, optional): Video extension (.mp4 or .avi). Defaults to '.mp4'.

    Returns:
        projconfigfile (str): Path to the new vame project config file.
    """

    date = dt.today()
    month = date.strftime("%B")
    day = date.day
    year = date.year
    d = str(month[0:3]+str(day))
    date = dt.today().strftime('%Y-%m-%d')

    wd = Path(working_directory).resolve()
    project_name = '{pn}-{date}'.format(pn=project, date=d+'-'+str(year))

    project_path = wd / project_name
    if project_path.exists():
        logger.info('Project "{}" already exists!'.format(project_path))
        projconfigfile = os.path.join(str(project_path),'config.yaml')
        return projconfigfile

    video_path = project_path / 'videos'
    data_path = project_path / 'data'
    results_path = project_path / 'results'
    model_path = project_path / 'model'

    for p in [video_path, data_path, results_path, model_path]:
        p.mkdir(parents=True)
        logger.info('Created "{}"'.format(p))

    vids = []
    for i in videos:
        #Check if it is a folder
        if os.path.isdir(i):
            vids_in_dir = [os.path.join(i,vp) for vp in os.listdir(i) if videotype in vp]
            vids = vids + vids_in_dir
            if len(vids_in_dir)==0:
                logger.info(f"No videos found in {i}")
                logger.info(f"Perhaps change the videotype, which is currently set to: {videotype}")
            else:
                videos = vids
                logger.info(
                    f"{len(vids_in_dir)} videos from the directory {i} were added to the project."
                )
        else:
            if os.path.isfile(i):
                vids = vids + [i]
            videos = vids

    pose_estimations_paths = []
    for pose_estimation_path in poses_estimations:
        if os.path.isdir(pose_estimation_path):
            pose_estimation_files = [os.path.join(pose_estimation_path, p) for p in os.listdir(pose_estimation_path) if '.csv' in p]
            pose_estimations_paths.extend(pose_estimation_files)
        else:
            pose_estimations_paths.append(pose_estimation_path)

    videos = [Path(vp) for vp in videos]
    video_names = []
    dirs_data = [data_path/Path(i.stem) for i in videos]
    for p in dirs_data:
        """
        Creates directory under data
        """
        p.mkdir(parents = True, exist_ok = True)
        video_names.append(p.stem)

    dirs_results = [results_path/Path(i.stem) for i in videos]
    for p in dirs_results:
        """
        Creates directory under results
        """
        p.mkdir(parents = True, exist_ok = True)

    destinations = [video_path.joinpath(vp.name) for vp in videos]

    os.mkdir(str(project_path)+'/'+'videos/pose_estimation/')
    os.mkdir(str(project_path)+'/model/pretrained_model')

    logger.info("Copying the videos \n")
    for src, dst in zip(videos, destinations):
        shutil.copy(os.fspath(src),os.fspath(dst))

    logger.info("Copying pose estimation files\n")
    for src, dst in zip(pose_estimations_paths, [str(project_path)+'/videos/pose_estimation/'+Path(p).name for p in pose_estimations_paths]):
        logger.info(f'Copying {src} to {dst}')
        shutil.copy(os.fspath(src),os.fspath(dst))

    new_project = ProjectSchema(
        Project=str(project),
        project_path=str(project_path),
        video_sets=video_names
    )
    cfg_data = new_project.model_dump()

    projconfigfile=os.path.join(str(project_path), 'config.yaml')
    # Write dictionary to yaml  config file
    write_config(projconfigfile, cfg_data)

    vame_pipeline_default_schema = VAMEPipelineStatesSchema()
    vame_pipeline_default_schema_path = Path(project_path) / 'states/states.json'
    if not vame_pipeline_default_schema_path.parent.exists():
        vame_pipeline_default_schema_path.parent.mkdir(parents=True)
    with open(vame_pipeline_default_schema_path, 'w') as f:
        json.dump(vame_pipeline_default_schema.model_dump(), f, indent=4)

    logger.info('A VAME project has been created. \n')
    logger.info('Now its time to prepare your data for VAME. '
          'The first step is to move your pose .csv file (e.g. DeepLabCut .csv) into the '
          '//YOUR//VAME//PROJECT//videos//pose_estimation folder. From here you can call '
          'either the function vame.egocentric_alignment() or if your data is by design egocentric '
          'call vame.csv_to_numpy(). This will prepare the data in .csv into the right format to start '
          'working with VAME.')

    return projconfigfile



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0

The following code is adapted from:

DeepLabCut2.0 Toolbox (deeplabcut.org)
© A. & M. Mathis Labs
https://github.com/AlexEMG/DeepLabCut
Please see AUTHORS for contributors.
https://github.com/AlexEMG/DeepLabCut/blob/master/AUTHORS
Licensed under GNU Lesser General Public License v3.0
"""

import os, yaml
from pathlib import Path
import ruamel.yaml
from typing import Tuple


def create_config_template() -> Tuple[dict, ruamel.yaml.YAML]:
    """
    Creates a template for the config.yaml file.

    Returns:
        Tuple[dict, ruamel.yaml.YAML]: A tuple containing the template dictionary and the Ruamel YAML instance.
    """
    yaml_str = """\
# Project configurations
    Project:
    model_name:
    n_cluster:
    pose_confidence:
    \n
# Project path and videos
    project_path:
    video_sets:
    \n
# Data
    all_data:
    \n
# Creation of train set:
    egocentric_data:
    robust:
    iqr_factor:
    axis:
    savgol_filter:
    savgol_length:
    savgol_order:
    test_fraction:
    \n
# RNN model general hyperparameter:
    pretrained_model:
    pretrained_weights:
    num_features:
    batch_size:
    max_epochs:
    model_snapshot:
    model_convergence:
    transition_function:
    beta:
    beta_norm:
    zdims:
    learning_rate:
    time_window:
    prediction_decoder:
    prediction_steps:
    noise:
    scheduler:
    scheduler_step_size:
    scheduler_gamma:
    #Note the optimal scheduler threshold below can vary greatly (from .1-.0001) between experiments.
    #You are encouraged to read the torch.optim.ReduceLROnPlateau docs to understand the threshold to use.
    scheduler_threshold:
    softplus:
    \n
# Segmentation:
    parametrization:
    hmm_trained: False
    load_data:
    individual_parametrization:
    random_state_kmeans:
    n_init_kmeans:
    \n
# Video writer:
    length_of_motif_video:
    \n
# UMAP parameter:
    min_dist:
    n_neighbors:
    random_state:
    num_points:
    \n
# ONLY CHANGE ANYTHING BELOW IF YOU ARE FAMILIAR WITH RNN MODELS
# RNN encoder hyperparamter:
    hidden_size_layer_1:
    hidden_size_layer_2:
    dropout_encoder:
    \n
# RNN reconstruction hyperparameter:
    hidden_size_rec:
    dropout_rec:
    n_layers:
    \n
# RNN prediction hyperparamter:
    hidden_size_pred:
    dropout_pred:
    \n
# RNN loss hyperparameter:
    mse_reconstruction_reduction:
    mse_prediction_reduction:
    kmeans_loss:
    kmeans_lambda:
    anneal_function:
    kl_start:
    annealtime:
    """
    ruamelFile = ruamel.yaml.YAML()
    cfg_file = ruamelFile.load(yaml_str)
    return(cfg_file,ruamelFile)


def read_config(configname: str) -> dict:
    """
    Reads structured config file defining a project.

    Args:
        configname (str): Path to the config file.

    Returns:
        dict: The contents of the config file as a dictionary.
    """
    ruamelFile = ruamel.yaml.YAML()
    path = Path(configname)
    if os.path.exists(path):
        try:
            with open(path, "r") as f:
                cfg = ruamelFile.load(f)
                curr_dir = os.path.dirname(configname)
                if cfg["project_path"] != curr_dir:
                    cfg["project_path"] = curr_dir
                    write_config(configname, cfg)
        except Exception as err:
            if len(err.args) > 2:
                if (
                    err.args[2]
                    == "could not determine a constructor for the tag '!!python/tuple'"
                ):
                    with open(path, "r") as ymlfile:
                        cfg = yaml.load(ymlfile, Loader=yaml.SafeLoader)
                        write_config(configname, cfg)
                else:
                    raise

    else:
        raise FileNotFoundError(
            "Config file is not found. Please make sure that the file exists and/or that you passed the path of the config file correctly!"
        )
    return cfg


def write_config(configname: str, cfg: dict) -> None:
    """
    Write structured config file.

    Args:
        configname (str): Path to the config file.
        cfg (dict): Dictionary containing the config data.
    """
    with open(configname, 'w') as cf:
        ruamelFile = ruamel.yaml.YAML()
        cfg_file,ruamelFile = create_config_template()
        for key in cfg.keys():
            cfg_file[key]=cfg[key]

        ruamelFile.dump(cfg_file, cf)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
from pathlib import Path
import numpy as np
import cv2 as cv
import tqdm
from typing import Union
from vame.util.auxiliary import read_config
from vame.schemas.states import save_state, MotifVideosFunctionSchema, CommunityVideosFunctionSchema
import imageio
from vame.logging.logger import VameLogger, TqdmToLogger
from vame.schemas.project import Parametrizations


logger_config = VameLogger(__name__)
logger = logger_config.logger


def get_cluster_vid(
    cfg: dict,
    path_to_file: str,
    file: str,
    n_cluster: int,
    videoType: str,
    flag: str,
    param: Parametrizations,
    output_video_type: str = ".mp4",
    tqdm_logger_stream: TqdmToLogger | None = None,
) -> None:
    """
    Generate cluster videos.

    Args:
        cfg (dict): Configuration parameters.
        path_to_file (str): Path to the file.
        file (str): Name of the file.
        n_cluster (int): Number of clusters.
        videoType (str): Type of input video.
        flag (str): Flag indicating the type of video (motif or community).

    Returns:
        None - Generate cluster videos and save them to fs on project folder.
    """
    if output_video_type not in ['.mp4', '.avi']:
        raise ValueError("Output video type must be either '.avi' or '.mp4'.")


    if flag == "motif":
        logger.info("Motif videos getting created for "+file+" ...")
        labels = np.load(os.path.join(path_to_file,str(n_cluster)+'_' + param + '_label_'+file+'.npy'))
    if flag == "community":
        logger.info("Community videos getting created for "+file+" ...")
        labels = np.load(os.path.join(path_to_file,"community",'community_label_'+file+'.npy'))
    capture = cv.VideoCapture(os.path.join(cfg['project_path'],"videos",file+videoType))

    if capture.isOpened():
        width  = capture.get(cv.CAP_PROP_FRAME_WIDTH)
        height = capture.get(cv.CAP_PROP_FRAME_HEIGHT)
        fps = 25#capture.get(cv.CAP_PROP_FPS)

    cluster_start = cfg['time_window'] / 2

    for cluster in range(n_cluster):
        logger.info('Cluster: %d' %(cluster))
        cluster_lbl = np.where(labels == cluster)
        cluster_lbl = cluster_lbl[0]
        if not cluster_lbl.size:
            logger.info('Cluster is empty')
            continue

        if flag == "motif":
            output = os.path.join(path_to_file,"cluster_videos",file+f'-motif_%d{output_video_type}' %cluster)
        if flag == "community":
            output = os.path.join(path_to_file,"community_videos",file+f'-community_%d{output_video_type}' %cluster)

        if output_video_type == '.avi':
            codec = cv.VideoWriter_fourcc("M", "J", "P", "G")
            video_writer = cv.VideoWriter(output, codec, fps, (int(width), int(height)))
        elif output_video_type == '.mp4':
            video_writer = imageio.get_writer(output, fps=fps, codec='h264', macro_block_size=None)


        if len(cluster_lbl) < cfg['length_of_motif_video']:
            vid_length = len(cluster_lbl)
        else:
            vid_length = cfg['length_of_motif_video']

        for num in tqdm.tqdm(range(vid_length), file=tqdm_logger_stream):
            idx = cluster_lbl[num]
            capture.set(1,idx+cluster_start)
            ret, frame = capture.read()
            if output_video_type == '.avi':
                video_writer.write(frame)
            elif output_video_type == '.mp4':
                video_writer.append_data(frame)
        if output_video_type == '.avi':
            video_writer.release()
        elif output_video_type == '.mp4':
            video_writer.close()
    capture.release()

@save_state(model=MotifVideosFunctionSchema)
def motif_videos(
    config: Union[str, Path],
    parametrization: Parametrizations,
    videoType: str = '.mp4',
    output_video_type: str = '.mp4',
    save_logs: bool = False
) -> None:
    """
    Generate motif videos.

    Args:
        config (Union[str, Path]): Path to the configuration file.
        videoType (str, optional): Type of video. Default is '.mp4'.
        output_video_type (str, optional): Type of output video. Default is '.mp4'.

    Returns:
        None - Generate motif videos and save them to filesystem on project cluster_videos folder.
    """
    try:
        tqdm_logger_stream = None
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        parametrizations = cfg['parametrizations']

        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'motif_videos.log'
            logger_config.add_file_handler(log_path)
            tqdm_logger_stream = TqdmToLogger(logger=logger)
        model_name = cfg['model_name']
        n_cluster = cfg['n_cluster']
        flag = 'motif'


        logger.info(f"Creating motif videos for parametrization: {parametrization}...")
        files = []
        if cfg['all_data'] == 'No':
            all_flag = input("Do you want to write motif videos for your entire dataset? \n"
                        "If you only want to use a specific dataset type filename: \n"
                        "yes/no/filename ")
        else:
            all_flag = 'yes'

        if all_flag == 'yes' or all_flag == 'Yes':
            for file in cfg['video_sets']:
                files.append(file)

        elif all_flag == 'no' or all_flag == 'No':
            for file in cfg['video_sets']:
                use_file = input("Do you want to quantify " + file + "? yes/no: ")
                if use_file == 'yes':
                    files.append(file)
                if use_file == 'no':
                    continue
        else:
            files.append(all_flag)

        logger.info("Cluster size is: %d " %n_cluster)
        for file in files:
            path_to_file=os.path.join(cfg['project_path'],"results",file,model_name, parametrization+'-'+str(n_cluster),"")
            if not os.path.exists(os.path.join(path_to_file,"cluster_videos")):
                os.mkdir(os.path.join(path_to_file,"cluster_videos"))

            get_cluster_vid(cfg, path_to_file, file, n_cluster, videoType, flag, parametrization, output_video_type=output_video_type, tqdm_logger_stream=tqdm_logger_stream)
        logger.info("All videos have been created!")
    except Exception as e:
        logger.exception(f"Error in motif_videos: {e}")
        raise e
    finally:
        logger_config.remove_file_handler()


@save_state(model=CommunityVideosFunctionSchema)
def community_videos(
    config: Union[str, Path],
    parametrization: Parametrizations,
    videoType: str = '.mp4',
    save_logs: bool = False,
    output_video_type: str = '.mp4'
) -> None:
    """
    Generate community videos.

    Args:
        config (Union[str, Path]): Path to the configuration file.
        videoType (str, optional): Type of video. Default is '.mp4'.

    Returns:
        None - Generate community videos and save them to filesystem on project community_videos folder.
    """
    try:
        tqdm_logger_stream = None
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        parametrizations = cfg['parametrizations']

        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'community_videos.log'
            logger_config.add_file_handler(log_path)
            tqdm_logger_stream = TqdmToLogger(logger=logger)
        model_name = cfg['model_name']
        n_cluster = cfg['n_cluster']
        flag = 'community'

        files = []
        if cfg['all_data'] == 'No':
            all_flag = input("Do you want to write motif videos for your entire dataset? \n"
                        "If you only want to use a specific dataset type filename: \n"
                        "yes/no/filename ")
        else:
            all_flag = 'yes'

        if all_flag == 'yes' or all_flag == 'Yes':
            for file in cfg['video_sets']:
                files.append(file)

        elif all_flag == 'no' or all_flag == 'No':
            for file in cfg['video_sets']:
                use_file = input("Do you want to quantify " + file + "? yes/no: ")
                if use_file == 'yes':
                    files.append(file)
                if use_file == 'no':
                    continue
        else:
            files.append(all_flag)

        logger.info("Cluster size is: %d " %n_cluster)
        for file in files:
            path_to_file=os.path.join(cfg['project_path'],"results",file,model_name,parametrization+'-'+str(n_cluster),"")
            if not os.path.exists(os.path.join(path_to_file,"community_videos")):
                os.mkdir(os.path.join(path_to_file,"community_videos"))

            get_cluster_vid(cfg, path_to_file, file, n_cluster, videoType, flag, parametrization, tqdm_logger_stream=tqdm_logger_stream, output_video_type=output_video_type)

        logger.info("All videos have been created!")

    except Exception as e:
        logger.exception(f"Error in community_videos: {e}")
        raise e
    finally:
        logger_config.remove_file_handler()


"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & J. Kürsch & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import cv2 as cv
import numpy as np
import pandas as pd
import tqdm
from typing import Tuple, List
from vame.logging.logger import VameLogger, TqdmToLogger
from pathlib import Path
from vame.util.auxiliary import read_config
from vame.schemas.states import EgocentricAlignmentFunctionSchema, save_state
from vame.util.data_manipulation import (
    interpol_first_rows_nans,
    crop_and_flip,
    background
)


logger_config = VameLogger(__name__)
logger = logger_config.logger

def align_mouse(
    path_to_file: str,
    filename: str,
    video_format: str,
    crop_size: Tuple[int, int],
    pose_list: List[np.ndarray],
    pose_ref_index: Tuple[int, int],
    confidence: float,
    pose_flip_ref: Tuple[int, int],
    bg: np.ndarray,
    frame_count: int,
    use_video: bool = True,
    tqdm_stream: TqdmToLogger = None
) -> Tuple[List[np.ndarray],List[List[np.ndarray]], np.ndarray]:
    """
    Align the mouse in the video frames.

    Args:
        path_to_file (str): Path to the file directory.
        filename (str): Name of the video file without the format.
        video_format (str): Format of the video file.
        crop_size (Tuple[int, int]): Size to crop the video frames.
        pose_list (List[np.ndarray]): List of pose coordinates.
        pose_ref_index (Tuple[int, int]): Pose reference indices.
        confidence (float): Pose confidence threshold.
        pose_flip_ref (Tuple[int, int]): Reference indices for flipping.
        bg (np.ndarray): Background image.
        frame_count (int): Number of frames to align.
        use_video (bool, optional): bool if video should be cropped or DLC points only. Defaults to True.

    Returns:
        Tuple[List[np.ndarray], List[List[np.ndarray]], np.ndarray]: List of aligned images, list of aligned DLC points, and time series data.
    """

    images = []
    points = []

    for i in pose_list:
        for j in i:
            if j[2] <= confidence:
                j[0],j[1] = np.nan, np.nan


    for i in pose_list:
        i = interpol_first_rows_nans(i)

    if use_video:
        capture = cv.VideoCapture(os.path.join(path_to_file,'videos',filename+video_format))

        if not capture.isOpened():
            raise Exception("Unable to open video file: {0}".format(os.path.join(path_to_file,'videos',filename+video_format)))

    for idx in tqdm.tqdm(range(frame_count), disable=not True, file=tqdm_stream, desc='Align frames'):

        if use_video:
            #Read frame
            try:
                ret, frame = capture.read()
                frame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
                frame = frame - bg
                frame[frame <= 0] = 0
            except Exception:
                logger.info("Couldn't find a frame in capture.read(). #Frame: %d" %idx)
                continue
        else:
            frame=np.zeros((1,1))

        #Read coordinates and add border
        pose_list_bordered = []

        for i in pose_list:
            pose_list_bordered.append((int(i[idx][0]+crop_size[0]),int(i[idx][1]+crop_size[1])))

        img = cv.copyMakeBorder(frame, crop_size[1], crop_size[1], crop_size[0], crop_size[0], cv.BORDER_CONSTANT, 0)


        coord_center = []
        punkte = []

        for i in pose_ref_index:
            coord = []

            coord.append(pose_list_bordered[i][0]) # changed from pose_list_bordered[i][0] 2/28/2024 PN
            coord.append(pose_list_bordered[i][1]) # changed from pose_list_bordered[i][1] 2/28/2024 PN

            punkte.append(coord)


        # coord_center.append(pose_list_bordered[5][0]-5)
        # coord_center.append(pose_list_bordered[5][0]+5)

        # coord_center = [coord_center]
        punkte = [punkte]

        # coord_center = np.asarray(coord_center)
        punkte = np.asarray(punkte)

        #calculate minimal rectangle around snout and tail
        rect = cv.minAreaRect(punkte)
        # rect_belly = cv.minAreaRect(coord_center)

        # center_belly, size_belly, theta_belly = rect_belly

        #change size in rect tuple structure to be equal to crop_size
        lst = list(rect)
        lst[1] = crop_size
        # lst[0] = center_belly
        rect = tuple(lst)

        center, size, theta = rect

        # lst2 = list(rect)
        # lst2[0][0] = center[0] - size[0]//2
        # lst2[0][1] = center[1] - size[1]//2

        # rect = tuple(lst2)

        # center[0] -= size[0]//2
        # center[1] -= size[0]//2 # added this shift to change center to belly 2/28/2024

        #crop image
        out, shifted_points = crop_and_flip(rect, img,pose_list_bordered,pose_flip_ref)

        if use_video: #for memory optimization, just save images when video is used.
            images.append(out)
        points.append(shifted_points)

    if use_video:
        capture.release()

    time_series = np.zeros((len(pose_list)*2,frame_count))
    for i in range(frame_count):
        idx = 0
        for j in range(len(pose_list)):
            time_series[idx:idx+2,i] = points[i][j]
            idx += 2

    return images, points, time_series


def play_aligned_video(a: List[np.ndarray], n: List[List[np.ndarray]], frame_count: int) -> None:
    """
    Play the aligned video.

    Args:
        a (List[np.ndarray]): List of aligned images.
        n (List[List[np.ndarray]]): List of aligned DLC points.
        frame_count (int): Number of frames in the video.
    """
    colors = [(255,0,0),(0,255,0),(0,0,255),(255,255,0),(255,0,255),(0,255,255),(0,0,0),(255,255,255)]

    for i in range(frame_count):
        # Capture frame-by-frame
        ret, frame = True,a[i]
        if ret == True:

          # Display the resulting frame
          frame = cv.cvtColor(frame.astype('uint8')*255, cv.COLOR_GRAY2BGR)
          im_color = cv.applyColorMap(frame, cv.COLORMAP_JET)

          for c,j in enumerate(n[i]):
              cv.circle(im_color,(j[0], j[1]), 5, colors[c], -1)

          cv.imshow('Frame',im_color)

          # Press Q on keyboard to  exit
          if cv.waitKey(25) & 0xFF == ord('q'):
            break

        # Break the loop
        else:
            break
    cv.destroyAllWindows()


def alignment(
    path_to_file: str,
    filename: str,
    pose_ref_index: List[int],
    video_format: str,
    crop_size: Tuple[int, int],
    confidence: float,
    use_video: bool = False,
    check_video: bool = False,
    tqdm_stream: TqdmToLogger = None
) -> Tuple[np.ndarray, List[np.ndarray]]:
    """
    Perform alignment of egocentric data.

    Args:
        path_to_file (str): Path to the file directory.
        filename (str): Name of the video file without the format.
        pose_ref_index (List[int]): Pose reference indices.
        video_format (str): Format of the video file.
        crop_size (Tuple[int, int]): Size to crop the video frames.
        confidence (float): Pose confidence threshold.
        use_video (bool, optional): Whether to use video for alignment. Defaults to False.
        check_video (bool, optional): Whether to check the aligned video. Defaults to False.

    Returns:
        Tuple[np.ndarray, List[np.ndarray]]: Aligned time series data and list of aligned frames.
    """

    #read out data
    data = pd.read_csv(os.path.join(path_to_file,'videos','pose_estimation',filename+'.csv'), skiprows = 2)
    data_mat = pd.DataFrame.to_numpy(data)
    data_mat = data_mat[:,1:]

    # get the coordinates for alignment from data table
    pose_list = []

    for i in range(int(data_mat.shape[1]/3)):
        pose_list.append(data_mat[:,i*3:(i+1)*3])

    #list of reference coordinate indices for alignment
    #0: snout, 1: forehand_left, 2: forehand_right,
    #3: hindleft, 4: hindright, 5: tail

    pose_ref_index = pose_ref_index

    #list of 2 reference coordinate indices for avoiding flipping
    pose_flip_ref = pose_ref_index

    if use_video:
        #compute background
        bg = background(path_to_file,filename,video_format, save_background=False)
        capture = cv.VideoCapture(os.path.join(path_to_file,'videos',filename+video_format))
        if not capture.isOpened():
            raise Exception("Unable to open video file: {0}".format(os.path.join(path_to_file,'videos',filename+video_format)))

        frame_count = int(capture.get(cv.CAP_PROP_FRAME_COUNT))
        capture.release()
    else:
        bg = 0
        frame_count = len(data) # Change this to an abitrary number if you first want to test the code


    frames, n, time_series = align_mouse(
        path_to_file,
        filename,
        video_format,
        crop_size,
        pose_list,
        pose_ref_index,
        confidence,
        pose_flip_ref,
        bg,
        frame_count,
        use_video,
        tqdm_stream,
    )

    if check_video:
        play_aligned_video(frames, n, frame_count)

    return time_series, frames


@save_state(model=EgocentricAlignmentFunctionSchema)
def egocentric_alignment(
    config: str,
    pose_ref_index: list = [5,6],
    crop_size: tuple = (300,300),
    use_video: bool = False,
    video_format: str = '.mp4',
    check_video: bool = False,
    save_logs: bool = False
) -> None:
    """Aligns egocentric data for VAME training

    Args:
        config (str): Path for the project config file.
        pose_ref_index (list, optional): Pose reference index to be used to align. Defaults to [5,6].
        crop_size (tuple, optional): Size to crop the video. Defaults to (300,300).
        use_video (bool, optional): Weather to use video to do the post alignment. Defaults to False. # TODO check what to put in this docstring
        video_format (str, optional): Video format, can be .mp4 or .avi. Defaults to '.mp4'.
        check_video (bool, optional): Weather to check the video. Defaults to False.

    Raises:
        ValueError: If the config.yaml indicates that the data is not egocentric.
    """

    # pose_ref_index changed in this script from [0,5] to                                                                                                                                         [5,6] on 2/7/2024 PN
    """ Happy aligning """
    #config parameters

    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        tqdm_stream = None
        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'egocentric_alignment.log'
            logger_config.add_file_handler(log_path)
            tqdm_stream = TqdmToLogger(logger=logger)

        logger.info('Starting egocentric alignment')
        path_to_file = cfg['project_path']
        filename = cfg['video_sets']
        confidence = cfg['pose_confidence']
        num_features = cfg['num_features']
        video_format=video_format
        crop_size=crop_size

        y_shifted_indices = np.arange(0, num_features, 2)
        x_shifted_indices = np.arange(1, num_features, 2)
        belly_Y_ind = pose_ref_index[0] * 2
        belly_X_ind = (pose_ref_index[0] * 2) + 1

        if cfg['egocentric_data']:
            raise ValueError("The config.yaml indicates that the data is egocentric. Please check the parameter egocentric_data")

        # call function and save into your VAME data folder
        for file in filename:
            logger.info("Aligning data %s, Pose confidence value: %.2f" %(file, confidence))
            egocentric_time_series, frames = alignment(path_to_file, file, pose_ref_index, video_format, crop_size,
                                                    confidence, use_video=use_video, check_video=check_video, tqdm_stream=tqdm_stream)

            # Shifiting section added 2/29/2024 PN
            egocentric_time_series_shifted = egocentric_time_series
            belly_Y_shift = egocentric_time_series[belly_Y_ind,:]
            belly_X_shift = egocentric_time_series[belly_X_ind,:]

            egocentric_time_series_shifted[y_shifted_indices, :] -= belly_Y_shift
            egocentric_time_series_shifted[x_shifted_indices, :] -= belly_X_shift

            np.save(os.path.join(path_to_file,'data',file,file+'-PE-seq.npy'), egocentric_time_series_shifted) # save new shifted file
    #        np.save(os.path.join(path_to_file,'data/',file,"",file+'-PE-seq.npy', egocentric_time_series))

        logger.info("Your data is now ine right format and you can call vame.create_trainset()")
    except Exception as e:
        logger.exception(f"{e}")
        raise e
    finally:
        logger_config.remove_file_handler()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""
import sys
sys.dont_write_bytecode = True

from vame.util.auxiliary import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import torch
from torch import nn
import torch.utils.data as Data
from torch.autograd import Variable
from torch.optim.lr_scheduler import StepLR, ReduceLROnPlateau
import os
import numpy as np
from pathlib import Path
from typing import Tuple
from vame.util.auxiliary import read_config
from vame.model.dataloader import SEQUENCE_DATASET
from vame.model.rnn_model import RNN_VAE
from tqdm import tqdm
from vame.schemas.states import TrainModelFunctionSchema, save_state
from vame.logging.logger import VameLogger, TqdmToLogger


logger_config = VameLogger(__name__)
logger = logger_config.logger
tqdm_to_logger = TqdmToLogger(logger)

# make sure torch uses cuda for GPU computing
use_gpu = torch.cuda.is_available()
if use_gpu:
    print("Using CUDA")
    print('GPU active:',torch.cuda.is_available())
    print('GPU used:',torch.cuda.get_device_name(0))
else:
    torch.device("cpu")

def reconstruction_loss(x: torch.Tensor, x_tilde: torch.Tensor, reduction: str) -> torch.Tensor:
    """Compute the reconstruction loss between input and reconstructed data.

    Args:
        x (torch.Tensor): Input data tensor.
        x_tilde (torch.Tensor): Reconstructed data tensor.
        reduction (str): Type of reduction for the loss.

    Returns:
        torch.Tensor: Reconstruction loss.
    """
    mse_loss = nn.MSELoss(reduction=reduction)
    rec_loss = mse_loss(x_tilde,x)
    return rec_loss

def future_reconstruction_loss(x: torch.Tensor, x_tilde: torch.Tensor, reduction: str) -> torch.Tensor:
    """Compute the future reconstruction loss between input and predicted future data.

    Args:
        x (torch.Tensor): Input future data tensor.
        x_tilde (torch.Tensor): Reconstructed future data tensor.
        reduction (str): Type of reduction for the loss.

    Returns:
        torch.Tensor: Future reconstruction loss.
    """
    mse_loss = nn.MSELoss(reduction=reduction)
    rec_loss = mse_loss(x_tilde,x)
    return rec_loss

def cluster_loss(H: torch.Tensor, kloss: int, lmbda: float, batch_size: int) -> torch.Tensor:
    """Compute the cluster loss.

    Args:
        H (torch.Tensor): Latent representation tensor.
        kloss (int): Number of clusters.
        lmbda (float): Lambda value for the loss.
        batch_size (int): Size of the batch.

    Returns:
        torch.Tensor: Cluster loss.
    """
    gram_matrix = (H.T @ H) / batch_size
    _ ,sv_2, _ = torch.svd(gram_matrix)
    sv = torch.sqrt(sv_2[:kloss])
    loss = torch.sum(sv)
    return lmbda*loss


def kullback_leibler_loss(mu: torch.Tensor, logvar: torch.Tensor) -> torch.Tensor:
    """Compute the Kullback-Leibler divergence loss.
    see Appendix B from VAE paper: Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014 - https://arxiv.org/abs/1312.6114

    Formula: 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)

    Args:
        mu (torch.Tensor): Mean of the latent distribution.
        logvar (torch.Tensor): Log variance of the latent distribution.

    Returns:
        torch.Tensor: Kullback-Leibler divergence loss.
    """
    # I'm using torch.mean() here as the sum() version depends on the size of the latent vector
    KLD = -0.5 * torch.mean(1 + logvar - mu.pow(2) - logvar.exp())
    return KLD


def kl_annealing(epoch: int, kl_start: int, annealtime: int, function: str) -> float:
    """
    Anneal the Kullback-Leibler loss to let the model learn first the reconstruction of the data
    before the KL loss term gets introduced.

    Args:
        epoch (int): Current epoch number.
        kl_start (int): Epoch number to start annealing the loss.
        annealtime (int): Annealing time.
        function (str): Annealing function type.

    Returns:
        float: Annealed weight value for the loss.
    """
    if epoch > kl_start:
        if function == 'linear':
            new_weight = min(1, (epoch-kl_start)/(annealtime))

        elif function == 'sigmoid':
            new_weight = float(1/(1+np.exp(-0.9*(epoch-annealtime))))
        else:
            raise NotImplementedError('currently only "linear" and "sigmoid" are implemented')

        return new_weight

    else:
        new_weight = 0
        return new_weight


def gaussian(ins: torch.Tensor, is_training: bool, seq_len: int, std_n: float = 0.8) -> torch.Tensor:
    """Add Gaussian noise to the input data.

    Args:
        ins (torch.Tensor): Input data tensor.
        is_training (bool): Whether it is training mode.
        seq_len (int): Length of the sequence.
        std_n (float): Standard deviation for the Gaussian noise.

    Returns:
        torch.Tensor: Noisy input data tensor.
    """
    if is_training:
        emp_std = ins.std(1)*std_n
        emp_std = emp_std.unsqueeze(2).repeat(1, 1, seq_len)
        emp_std = emp_std.permute(0,2,1)
        noise = Variable(ins.data.new(ins.size()).normal_(0, 1))
        return ins + (noise*emp_std)
    return ins



def train(
    train_loader: Data.DataLoader,
    epoch: int,
    model: nn.Module,
    optimizer: torch.optim.Optimizer,
    anneal_function: str,
    BETA: float,
    kl_start: int,
    annealtime: int,
    seq_len: int,
    future_decoder: bool,
    future_steps: int,
    scheduler: torch.optim.lr_scheduler._LRScheduler,
    mse_red: str,
    mse_pred: str,
    kloss: int,
    klmbda: float,
    bsize: int,
    noise: bool
) -> Tuple[float, float, float, float, float, float]:
    """Train the model.

    Args:
        train_loader (DataLoader): Training data loader.
        epoch (int): Current epoch number.
        model (nn.Module): Model to be trained.
        optimizer (Optimizer): Optimizer for training.
        anneal_function (str): Annealing function type.
        BETA (float): Beta value for the loss.
        kl_start (int): Epoch number to start annealing the loss.
        annealtime (int): Annealing time.
        seq_len (int): Length of the sequence.
        future_decoder (bool): Whether a future decoder is used.
        future_steps (int): Number of future steps to predict.
        scheduler (lr_scheduler._LRScheduler): Learning rate scheduler.
        mse_red (str): Reduction type for MSE reconstruction loss.
        mse_pred (str): Reduction type for MSE prediction loss.
        kloss (int): Number of clusters for cluster loss.
        klmbda (float): Lambda value for cluster loss.
        bsize (int): Size of the batch.
        noise (bool): Whether to add Gaussian noise to the input.

    Returns:
        Tuple[float, float, float, float, float, float]: Kullback-Leibler weight, train loss, K-means loss, KL loss,
        MSE loss, future loss.
    """

    model.train() # toggle model to train mode
    train_loss = 0.0
    mse_loss = 0.0
    kullback_loss = 0.0
    kmeans_losses = 0.0
    fut_loss = 0.0
    loss = 0.0
    seq_len_half = int(seq_len / 2)

    for idx, data_item in enumerate(train_loader):
        data_item = Variable(data_item)
        data_item = data_item.permute(0,2,1)

        if use_gpu:
            data = data_item[:,:seq_len_half,:].type('torch.FloatTensor').cuda()
            fut = data_item[:,seq_len_half:seq_len_half+future_steps,:].type('torch.FloatTensor').cuda()
        else:
            data = data_item[:,:seq_len_half,:].type('torch.FloatTensor').to()
            fut = data_item[:,seq_len_half:seq_len_half+future_steps,:].type('torch.FloatTensor').to()
        if noise == True:
            data_gaussian = gaussian(data,True,seq_len_half)
        else:
            data_gaussian = data

        if future_decoder:
            data_tilde, future, latent, mu, logvar = model(data_gaussian)

            rec_loss = reconstruction_loss(data, data_tilde, mse_red)
            fut_rec_loss = future_reconstruction_loss(fut, future, mse_pred)
            kmeans_loss = cluster_loss(latent.T, kloss, klmbda, bsize)
            kl_loss = kullback_leibler_loss(mu, logvar)
            kl_weight = kl_annealing(epoch, kl_start, annealtime, anneal_function)
            loss = rec_loss + fut_rec_loss + BETA*kl_weight*kl_loss + kl_weight*kmeans_loss
            fut_loss += fut_rec_loss.item()

        else:
            data_tilde, latent, mu, logvar = model(data_gaussian)

            rec_loss = reconstruction_loss(data, data_tilde, mse_red)
            kl_loss = kullback_leibler_loss(mu, logvar)
            kmeans_loss = cluster_loss(latent.T, kloss, klmbda, bsize)
            kl_weight = kl_annealing(epoch, kl_start, annealtime, anneal_function)
            loss = rec_loss + BETA*kl_weight*kl_loss + kl_weight*kmeans_loss

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm = 5)

        train_loss += loss.item()
        mse_loss += rec_loss.item()
        kullback_loss += kl_loss.item()
        kmeans_losses += kmeans_loss.item()

        # if idx % 1000 == 0:
        #     print('Epoch: %d.  loss: %.4f' %(epoch, loss.item()))

    scheduler.step(loss) #be sure scheduler is called before optimizer in >1.1 pytorch

    if future_decoder:
        print('Train loss: {:.3f}, MSE-Loss: {:.3f}, MSE-Future-Loss {:.3f}, KL-Loss: {:.3f}, Kmeans-Loss: {:.3f}, weight: {:.2f}'.format(train_loss / idx,
              mse_loss /idx, fut_loss/idx, BETA*kl_weight*kullback_loss/idx, kl_weight*kmeans_losses/idx, kl_weight))
    else:
        print('Train loss: {:.3f}, MSE-Loss: {:.3f}, KL-Loss: {:.3f}, Kmeans-Loss: {:.3f}, weight: {:.2f}'.format(train_loss / idx,
              mse_loss /idx, BETA*kl_weight*kullback_loss/idx, kl_weight*kmeans_losses/idx, kl_weight))

    return kl_weight, train_loss/idx, kl_weight*kmeans_losses/idx, kullback_loss/idx, mse_loss/idx, fut_loss/idx


def test(
    test_loader: Data.DataLoader,
    model: nn.Module,
    BETA: float,
    kl_weight: float,
    seq_len: int,
    mse_red: str,
    kloss: str,
    klmbda: float,
    future_decoder: bool,
    bsize: int
) -> Tuple[float, float, float]:
    """
    Evaluate the model on the test dataset.

    Args:
        test_loader (DataLoader): DataLoader for the test dataset.
        epoch (int, deprecated): Current epoch number.
        model (nn.Module): The trained model.
        optimizer (Optimizer, deprecated): The optimizer used for training.
        BETA (float): Beta value for the VAE loss.
        kl_weight (float): Weighting factor for the KL divergence loss.
        seq_len (int): Length of the sequence.
        mse_red (str): Reduction method for the MSE loss.
        kloss (str): Loss function for K-means clustering.
        klmbda (float): Lambda value for K-means loss.
        future_decoder (bool): Flag indicating whether to use a future decoder.
        bsize (int): Batch size.

    Returns:
        Tuple[float, float, float]: Tuple containing MSE loss per item, total test loss per item,
        and K-means loss weighted by the kl_weight.
    """

    model.eval() # toggle model to inference mode
    test_loss = 0.0
    mse_loss = 0.0
    kullback_loss = 0.0
    kmeans_losses = 0.0
    loss = 0.0
    seq_len_half = int(seq_len / 2)

    with torch.no_grad():
        for idx, data_item in enumerate(test_loader):
            # we're only going to infer, so no autograd at all required
            data_item = Variable(data_item)
            data_item = data_item.permute(0,2,1)
            if use_gpu:
                data = data_item[:,:seq_len_half,:].type('torch.FloatTensor').cuda()
            else:
                data = data_item[:,:seq_len_half,:].type('torch.FloatTensor').to()

            if future_decoder:
                recon_images, _, latent, mu, logvar = model(data)
                rec_loss = reconstruction_loss(data, recon_images, mse_red)
                kl_loss = kullback_leibler_loss(mu, logvar)
                kmeans_loss = cluster_loss(latent.T, kloss, klmbda, bsize)
                loss = rec_loss + BETA*kl_weight*kl_loss+ kl_weight*kmeans_loss

            else:
                recon_images, latent, mu, logvar = model(data)
                rec_loss = reconstruction_loss(data, recon_images, mse_red)
                kl_loss = kullback_leibler_loss(mu, logvar)
                kmeans_loss = cluster_loss(latent.T, kloss, klmbda, bsize)
                loss = rec_loss + BETA*kl_weight*kl_loss + kl_weight*kmeans_loss

            # torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm = 5)

            test_loss += loss.item()
            mse_loss += rec_loss.item()
            kullback_loss += kl_loss.item()
            kmeans_losses += kmeans_loss

    print('Test loss: {:.3f}, MSE-Loss: {:.3f}, KL-Loss: {:.3f}, Kmeans-Loss: {:.3f}'.format(test_loss / idx,
          mse_loss /idx, BETA*kl_weight*kullback_loss/idx, kl_weight*kmeans_losses/idx))

    return mse_loss /idx, test_loss/idx, kl_weight*kmeans_losses


@save_state(model=TrainModelFunctionSchema)
def train_model(config: str, save_logs: bool = False) -> None:
    """Train Variational Autoencoder using the configuration file values.

    Args:
        config (str): Path to the configuration file.
    """
    try:
        tqdm_logger_stream = None
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        if save_logs:
            tqdm_logger_stream = TqdmToLogger(logger)
            log_path = Path(cfg['project_path']) / 'logs' / 'train_model.log'
            logger_config.add_file_handler(log_path)

        model_name = cfg['model_name']
        pretrained_weights = cfg['pretrained_weights']
        pretrained_model = cfg['pretrained_model']
        fixed = cfg['egocentric_data']

        logger.info("Train Variational Autoencoder - model name: %s \n" %model_name)
        if not os.path.exists(os.path.join(cfg['project_path'],'model','best_model',"")):
            os.mkdir(os.path.join(cfg['project_path'],'model','best_model',""))
            os.mkdir(os.path.join(cfg['project_path'],'model','best_model','snapshots',""))
            os.mkdir(os.path.join(cfg['project_path'],'model','model_losses',""))

        # make sure torch uses cuda for GPU computing
        use_gpu = torch.cuda.is_available()
        if use_gpu:
            logger.info("Using CUDA")
            logger.info("GPU active: {}".format(torch.cuda.is_available()))
            logger.info('GPU used: {}'.format(torch.cuda.get_device_name(0)))
        else:
            torch.device("cpu")
            logger.info("warning, a GPU was not found... proceeding with CPU (slow!) \n")
            #raise NotImplementedError('GPU Computing is required!')

        """ HYPERPARAMTERS """
        # General
        CUDA = use_gpu
        SEED = 19
        TRAIN_BATCH_SIZE = cfg['batch_size']
        TEST_BATCH_SIZE = int(cfg['batch_size']/4)
        EPOCHS = cfg['max_epochs']
        ZDIMS = cfg['zdims']
        BETA  = cfg['beta']
        SNAPSHOT = cfg['model_snapshot']
        LEARNING_RATE = cfg['learning_rate']
        NUM_FEATURES = cfg['num_features']
        if not fixed:
            NUM_FEATURES = NUM_FEATURES - 2
        TEMPORAL_WINDOW = cfg['time_window']*2
        FUTURE_DECODER = cfg['prediction_decoder']
        FUTURE_STEPS = cfg['prediction_steps']

        # RNN
        hidden_size_layer_1 = cfg['hidden_size_layer_1']
        hidden_size_layer_2 = cfg['hidden_size_layer_2']
        hidden_size_rec = cfg['hidden_size_rec']
        hidden_size_pred = cfg['hidden_size_pred']
        dropout_encoder = cfg['dropout_encoder']
        dropout_rec = cfg['dropout_rec']
        dropout_pred = cfg['dropout_pred']
        noise = cfg['noise']
        scheduler_step_size = cfg['scheduler_step_size']
        softplus = cfg['softplus']

        # Loss
        MSE_REC_REDUCTION = cfg['mse_reconstruction_reduction']
        MSE_PRED_REDUCTION = cfg['mse_prediction_reduction']
        KMEANS_LOSS = cfg['kmeans_loss']
        KMEANS_LAMBDA = cfg['kmeans_lambda']
        KL_START = cfg['kl_start']
        ANNEALTIME = cfg['annealtime']
        anneal_function = cfg['anneal_function']
        optimizer_scheduler = cfg['scheduler']

        BEST_LOSS = 999999
        convergence = 0
        logger.info('Latent Dimensions: %d, Time window: %d, Batch Size: %d, Beta: %d, lr: %.4f\n' %(ZDIMS, cfg['time_window'], TRAIN_BATCH_SIZE, BETA, LEARNING_RATE))

        # simple logging of diverse losses
        train_losses = []
        test_losses = []
        kmeans_losses = []
        kl_losses = []
        weight_values = []
        mse_losses = []
        fut_losses = []

        torch.manual_seed(SEED)
        RNN = RNN_VAE
        if CUDA:
            torch.cuda.manual_seed(SEED)
            model = RNN(TEMPORAL_WINDOW,ZDIMS,NUM_FEATURES,FUTURE_DECODER,FUTURE_STEPS, hidden_size_layer_1,
                            hidden_size_layer_2, hidden_size_rec, hidden_size_pred, dropout_encoder,
                            dropout_rec, dropout_pred, softplus).cuda()
        else: #cpu support ...
            torch.cuda.manual_seed(SEED)
            model = RNN(TEMPORAL_WINDOW,ZDIMS,NUM_FEATURES,FUTURE_DECODER,FUTURE_STEPS, hidden_size_layer_1,
                            hidden_size_layer_2, hidden_size_rec, hidden_size_pred, dropout_encoder,
                            dropout_rec, dropout_pred, softplus).to()

        if pretrained_weights:
            try:
                logger.info("Loading pretrained weights from model: %s\n" %os.path.join(cfg['project_path'],'model','best_model',pretrained_model+'_'+cfg['Project']+'.pkl'))
                model.load_state_dict(torch.load(os.path.join(cfg['project_path'],'model','best_model',pretrained_model+'_'+cfg['Project']+'.pkl')))
                KL_START = 0
                ANNEALTIME = 1
            except Exception:
                logger.info("No file found at %s\n" %os.path.join(cfg['project_path'],'model','best_model',pretrained_model+'_'+cfg['Project']+'.pkl'))
                try:
                    logger.info("Loading pretrained weights from %s\n" %pretrained_model)
                    model.load_state_dict(torch.load(pretrained_model))
                    KL_START = 0
                    ANNEALTIME = 1
                except Exception:
                    logger.error("Could not load pretrained model. Check file path in config.yaml.")

        """ DATASET """
        trainset = SEQUENCE_DATASET(os.path.join(cfg['project_path'],"data", "train",""), data='train_seq.npy', train=True, temporal_window=TEMPORAL_WINDOW)
        testset = SEQUENCE_DATASET(os.path.join(cfg['project_path'],"data", "train",""), data='test_seq.npy', train=False, temporal_window=TEMPORAL_WINDOW)

        train_loader = Data.DataLoader(trainset, batch_size=TRAIN_BATCH_SIZE, shuffle=True, drop_last=True)
        test_loader = Data.DataLoader(testset, batch_size=TEST_BATCH_SIZE, shuffle=True, drop_last=True)

        optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE, amsgrad=True)

        if optimizer_scheduler:
            logger.info('Scheduler step size: %d, Scheduler gamma: %.2f\n' %(scheduler_step_size, cfg['scheduler_gamma']))
            # Thanks to @alexcwsmith for the optimized scheduler contribution
            scheduler = ReduceLROnPlateau(optimizer, 'min', factor=cfg['scheduler_gamma'], patience=cfg['scheduler_step_size'], threshold=1e-3, threshold_mode='rel', verbose=True)
        else:
            scheduler = StepLR(optimizer, step_size=scheduler_step_size, gamma=1, last_epoch=-1)

        logger.info("Start training... ")
        for epoch in tqdm(range(1,EPOCHS), desc="Training Model", unit="epoch", file=tqdm_logger_stream):
            #print("Epoch: %d" %epoch)
            weight, train_loss, km_loss, kl_loss, mse_loss, fut_loss = train(train_loader, epoch, model, optimizer,
                                                                            anneal_function, BETA, KL_START,
                                                                            ANNEALTIME, TEMPORAL_WINDOW, FUTURE_DECODER,
                                                                            FUTURE_STEPS, scheduler, MSE_REC_REDUCTION,
                                                                            MSE_PRED_REDUCTION, KMEANS_LOSS, KMEANS_LAMBDA,
                                                                            TRAIN_BATCH_SIZE, noise)
            current_loss, test_loss, test_list = test(test_loader, model,
                                                    BETA, weight, TEMPORAL_WINDOW, MSE_REC_REDUCTION,
                                                    KMEANS_LOSS, KMEANS_LAMBDA, FUTURE_DECODER, TEST_BATCH_SIZE)

            # logging losses
            train_losses.append(train_loss)
            test_losses.append(test_loss)
            kmeans_losses.append(km_loss)
            kl_losses.append(kl_loss)
            weight_values.append(weight)
            mse_losses.append(mse_loss)
            fut_losses.append(fut_loss)

            # save best model
            if weight > 0.99 and current_loss <= BEST_LOSS:
                BEST_LOSS = current_loss
                logger.info("Saving model!")

                if use_gpu:
                    torch.save(model.state_dict(), os.path.join(cfg['project_path'],"model", "best_model",model_name+'_'+cfg['Project']+'.pkl'))

                else:
                    torch.save(model.state_dict(), os.path.join(cfg['project_path'],"model", "best_model",model_name+'_'+cfg['Project']+'.pkl'))

                convergence = 0
            else:
                convergence += 1

            if epoch % SNAPSHOT == 0:
                logger.info("Saving model snapshot!\n")
                torch.save(model.state_dict(), os.path.join(cfg['project_path'],'model','best_model','snapshots',model_name+'_'+cfg['Project']+'_epoch_'+str(epoch)+'.pkl'))

            if convergence > cfg['model_convergence']:
                logger.info('Finished training...')
                logger.info('Model converged. Please check your model with vame.evaluate_model(). \n'
                    'You can also re-run vame.trainmodel() to further improve your model. \n'
                    'Make sure to set _pretrained_weights_ in your config.yaml to "true" \n'
                    'and plug your current model name into _pretrained_model_. \n'
                    'Hint: Set "model_convergence" in your config.yaml to a higher value. \n'
                    '\n'
                    'Next: \n'
                    'Use vame.pose_segmentation() to identify behavioral motifs in your dataset!')
                break

            # save logged losses
            np.save(os.path.join(cfg['project_path'],'model','model_losses','train_losses_'+model_name), train_losses)
            np.save(os.path.join(cfg['project_path'],'model','model_losses','test_losses_'+model_name), test_losses)
            np.save(os.path.join(cfg['project_path'],'model','model_losses','kmeans_losses_'+model_name), kmeans_losses)
            np.save(os.path.join(cfg['project_path'],'model','model_losses','kl_losses_'+model_name), kl_losses)
            np.save(os.path.join(cfg['project_path'],'model','model_losses','weight_values_'+model_name), weight_values)
            np.save(os.path.join(cfg['project_path'],'model','model_losses','mse_train_losses_'+model_name), mse_losses)
            np.save(os.path.join(cfg['project_path'],'model','model_losses','mse_test_losses_'+model_name), current_loss)
            np.save(os.path.join(cfg['project_path'],'model','model_losses','fut_losses_'+model_name), fut_losses)

            print("\n")

        if convergence < cfg['model_convergence']:
            logger.info('Finished training...')
            logger.info('Model seems to have not reached convergence. You may want to check your model \n'
                'with vame.evaluate_model(). If your satisfied you can continue. \n'
                'Use vame.pose_segmentation() to identify behavioral motifs! \n'
                'OPTIONAL: You can re-run vame.train_model() to improve performance.')

    except Exception as e:
        logger.exception(f"An error occurred: {e}")
        raise e
    finally:
        logger_config.remove_file_handler()



# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0

The Model is partially adapted from the Timeseries Clustering repository developed by Tejas Lodaya:
https://github.com/tejaslodaya/timeseries-clustering-vae/blob/master/vrae/vrae.py
"""


import torch
from torch import nn


# NEW MODEL WITH SMALL ALTERATIONS
""" MODEL  """

class Encoder(nn.Module):
    """Encoder module of the Variational Autoencoder."""
    def __init__(self, NUM_FEATURES: int, hidden_size_layer_1: int, hidden_size_layer_2: int, dropout_encoder: float):
        """
        Initialize the Encoder module.

        Args:
            NUM_FEATURES (int): Number of input features.
            hidden_size_layer_1 (int): Size of the first hidden layer.
            hidden_size_layer_2 (int): Size of the second hidden layer.
            dropout_encoder (float): Dropout rate for regularization.
        """
        super(Encoder, self).__init__()

        self.input_size = NUM_FEATURES
        self.hidden_size = hidden_size_layer_1
        self.hidden_size_2 = hidden_size_layer_2
        self.n_layers  = 2
        self.dropout   = dropout_encoder
        self.bidirectional = True

        self.encoder_rnn = nn.GRU(input_size=self.input_size, hidden_size=self.hidden_size, num_layers=self.n_layers,
                            bias=True, batch_first=True, dropout=self.dropout, bidirectional=self.bidirectional)#UNRELEASED!


        self.hidden_factor = (2 if self.bidirectional else 1) * self.n_layers

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the Encoder module.

        Args:
            inputs (torch.Tensor): Input tensor of shape (batch_size, sequence_length, num_features).

        Returns:
            torch.Tensor: Encoded representation tensor of shape (batch_size, hidden_size_layer_1 * 4).
        """
        outputs_1, hidden_1 = self.encoder_rnn(inputs)#UNRELEASED!

        hidden = torch.cat((hidden_1[0,...], hidden_1[1,...], hidden_1[2,...], hidden_1[3,...]),1)

        return hidden


class Lambda(nn.Module):
    """Lambda module for computing the latent space parameters."""
    def __init__(self, ZDIMS: int, hidden_size_layer_1: int, softplus: bool):
        """
        Initialize the Lambda module.

        Args:
            ZDIMS (int): Size of the latent space.
            hidden_size_layer_1 (int): Size of the first hidden layer.
            hidden_size_layer_2 (int, deprecated): Size of the second hidden layer.
            softplus (bool): Whether to use softplus activation for logvar.
        """
        super(Lambda, self).__init__()

        self.hid_dim = hidden_size_layer_1*4
        self.latent_length = ZDIMS
        self.softplus = softplus

        self.hidden_to_mean = nn.Linear(self.hid_dim, self.latent_length)
        self.hidden_to_logvar = nn.Linear(self.hid_dim, self.latent_length)

        if self.softplus == True:
            print("Using a softplus activation to ensures that the variance is parameterized as non-negative and activated by a smooth function")
            self.softplus_fn = nn.Softplus()

    def forward(self, hidden: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Forward pass of the Lambda module.

        Args:
            hidden (torch.Tensor): Hidden representation tensor of shape (batch_size, hidden_size_layer_1 * 4).

        Returns:
            tuple[torch.Tensor, torch.Tensor, torch.Tensor]: Latent space tensor, mean tensor, logvar tensor.
        """
        self.mean = self.hidden_to_mean(hidden)
        if self.softplus == True:
            self.logvar = self.softplus_fn(self.hidden_to_logvar(hidden))
        else:
            self.logvar = self.hidden_to_logvar(hidden)

        if self.training:
            std = torch.exp(0.5 * self.logvar)
            eps = torch.randn_like(std)
            return eps.mul(std).add_(self.mean), self.mean, self.logvar
        else:
            return self.mean, self.mean, self.logvar


class Decoder(nn.Module):
    """Decoder module of the Variational Autoencoder."""
    def __init__(self, TEMPORAL_WINDOW: int, ZDIMS: int, NUM_FEATURES: int, hidden_size_rec: int, dropout_rec: float):
        """
        Initialize the Decoder module.

        Args:
            TEMPORAL_WINDOW (int): Size of the temporal window.
            ZDIMS (int): Size of the latent space.
            NUM_FEATURES (int): Number of input features.
            hidden_size_rec (int): Size of the recurrent hidden layer.
            dropout_rec (float): Dropout rate for regularization.
        """
        super(Decoder,self).__init__()

        self.num_features = NUM_FEATURES
        self.sequence_length = TEMPORAL_WINDOW
        self.hidden_size = hidden_size_rec
        self.latent_length = ZDIMS
        self.n_layers  = 1
        self.dropout   = dropout_rec
        self.bidirectional = True

        self.rnn_rec = nn.GRU(self.latent_length, hidden_size=self.hidden_size, num_layers=self.n_layers,
                            bias=True, batch_first=True, dropout=self.dropout, bidirectional=self.bidirectional)

        self.hidden_factor = (2 if self.bidirectional else 1) * self.n_layers # NEW

        self.latent_to_hidden = nn.Linear(self.latent_length,self.hidden_size * self.hidden_factor) # NEW
        self.hidden_to_output = nn.Linear(self.hidden_size*(2 if self.bidirectional else 1), self.num_features)

    def forward(self, inputs: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the Decoder module.

        Args:
            inputs (torch.Tensor): Input tensor of shape (batch_size, seq_len, ZDIMS).
            z (torch.Tensor): Latent space tensor of shape (batch_size, ZDIMS).

        Returns:
            torch.Tensor: Decoded output tensor of shape (batch_size, seq_len, NUM_FEATURES).
        """
        batch_size = inputs.size(0) # NEW

        hidden = self.latent_to_hidden(z) # NEW

        hidden = hidden.view(self.hidden_factor, batch_size, self.hidden_size) # NEW

        decoder_output, _ = self.rnn_rec(inputs, hidden)
        prediction = self.hidden_to_output(decoder_output)

        return prediction


class Decoder_Future(nn.Module):
    """Decoder module for predicting future sequences."""
    def __init__(
        self,
        TEMPORAL_WINDOW: int,
        ZDIMS: int,
        NUM_FEATURES: int,
        FUTURE_STEPS: int,
        hidden_size_pred: int,
        dropout_pred: float
    ):
        """
        Initialize the Decoder_Future module.

        Args:
            TEMPORAL_WINDOW (int): Size of the temporal window.
            ZDIMS (int): Size of the latent space.
            NUM_FEATURES (int): Number of input features.
            FUTURE_STEPS (int): Number of future steps to predict.
            hidden_size_pred (int): Size of the prediction hidden layer.
            dropout_pred (float): Dropout rate for regularization.
        """
        super(Decoder_Future,self).__init__()

        self.num_features = NUM_FEATURES
        self.future_steps = FUTURE_STEPS
        self.sequence_length = TEMPORAL_WINDOW
        self.hidden_size = hidden_size_pred
        self.latent_length = ZDIMS
        self.n_layers  = 1
        self.dropout   = dropout_pred
        self.bidirectional = True

        self.rnn_pred = nn.GRU(self.latent_length, hidden_size=self.hidden_size, num_layers=self.n_layers,
                            bias=True, batch_first=True, dropout=self.dropout, bidirectional=self.bidirectional)

        self.hidden_factor = (2 if self.bidirectional else 1) * self.n_layers # NEW

        self.latent_to_hidden = nn.Linear(self.latent_length,self.hidden_size * self.hidden_factor)
        self.hidden_to_output = nn.Linear(self.hidden_size*2, self.num_features)

    def forward(self, inputs: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the Decoder_Future module.

        Args:
            inputs (torch.Tensor): Input tensor of shape (batch_size, seq_len, ZDIMS).
            z (torch.Tensor): Latent space tensor of shape (batch_size, ZDIMS).

        Returns:
            torch.Tensor: Predicted future tensor of shape (batch_size, FUTURE_STEPS, NUM_FEATURES).
        """
        batch_size = inputs.size(0)

        hidden = self.latent_to_hidden(z)
        hidden = hidden.view(self.hidden_factor, batch_size, self.hidden_size)

        inputs = inputs[:,:self.future_steps,:]
        decoder_output, _ = self.rnn_pred(inputs, hidden)

        prediction = self.hidden_to_output(decoder_output)

        return prediction


class RNN_VAE(nn.Module):
    """Variational Autoencoder module."""
    def __init__(
        self,
        TEMPORAL_WINDOW: int,
        ZDIMS: int,
        NUM_FEATURES: int,
        FUTURE_DECODER: bool,
        FUTURE_STEPS: int,
        hidden_size_layer_1: int,
        hidden_size_layer_2: int,
        hidden_size_rec: int,
        hidden_size_pred: int,
        dropout_encoder: float,
        dropout_rec: float,
        dropout_pred: float,
        softplus: bool
    ):

        """
        Initialize the VAE module.

        Args:
            TEMPORAL_WINDOW (int): Size of the temporal window.
            ZDIMS (int): Size of the latent space.
            NUM_FEATURES (int): Number of input features.
            FUTURE_DECODER (bool): Whether to include a future decoder.
            FUTURE_STEPS (int): Number of future steps to predict.
            hidden_size_layer_1 (int): Size of the first hidden layer.
            hidden_size_layer_2 (int): Size of the second hidden layer.
            hidden_size_rec (int): Size of the recurrent hidden layer.
            hidden_size_pred (int): Size of the prediction hidden layer.
            dropout_encoder (float): Dropout rate for encoder.

        """
        super(RNN_VAE,self).__init__()

        self.FUTURE_DECODER = FUTURE_DECODER
        self.seq_len = int(TEMPORAL_WINDOW / 2)
        self.encoder = Encoder(NUM_FEATURES, hidden_size_layer_1, hidden_size_layer_2, dropout_encoder)
        self.lmbda = Lambda(ZDIMS, hidden_size_layer_1, softplus)
        self.decoder = Decoder(self.seq_len,ZDIMS,NUM_FEATURES, hidden_size_rec, dropout_rec)
        if FUTURE_DECODER:
            self.decoder_future = Decoder_Future(self.seq_len,ZDIMS,NUM_FEATURES,FUTURE_STEPS, hidden_size_pred,
                                                 dropout_pred)

    def forward(self, seq: torch.Tensor) -> tuple:
        """Forward pass of the VAE.

        Args:
            seq (torch.Tensor): Input sequence tensor of shape (batch_size, seq_len, NUM_FEATURES).

        Returns:
            Tuple containing:
                - If FUTURE_DECODER is True:
                    - prediction (torch.Tensor): Reconstructed input sequence tensor.
                    - future (torch.Tensor): Predicted future sequence tensor.
                    - z (torch.Tensor): Latent representation tensor.
                    - mu (torch.Tensor): Mean of the latent distribution tensor.
                    - logvar (torch.Tensor): Log variance of the latent distribution tensor.
                - If FUTURE_DECODER is False:
                    - prediction (torch.Tensor): Reconstructed input sequence tensor.
                    - z (torch.Tensor): Latent representation tensor.
                    - mu (torch.Tensor): Mean of the latent distribution tensor.
                    - logvar (torch.Tensor): Log variance of the latent distribution tensor.
        """

        """ Encode input sequence """
        h_n = self.encoder(seq)

        """ Compute the latent state via reparametrization trick """
        z, mu, logvar = self.lmbda(h_n)
        ins = z.unsqueeze(2).repeat(1, 1, self.seq_len)
        ins = ins.permute(0,2,1)

        """ Predict the future of the sequence from the latent state"""
        prediction = self.decoder(ins, z)

        if self.FUTURE_DECODER:
            future = self.decoder_future(ins, z)
            return prediction, future, z, mu, logvar
        else:
            return prediction, z, mu, logvar

